"""A module designed for optimizing the geometry of ligands.

Index
-----
.. currentmodule:: CAT.attachment.ligand_opt
.. autosummary::
    init_ligand_opt
    _parse_overwrite
    read_data
    start_ligand_jobs
    optimize_ligand
    _ligand_to_db
    split_mol
    get_frag_size
    get_dihed
    set_dihed

API
---
.. autofunction:: init_ligand_opt
.. autofunction:: _parse_overwrite
.. autofunction:: read_data
.. autofunction:: start_ligand_jobs
.. autofunction:: optimize_ligand
.. autofunction:: _ligand_to_db
.. autofunction:: get_frag_size
.. autofunction:: split_bond
.. autofunction:: get_dihed
.. autofunction:: set_dihed

"""

import itertools
from types import MappingProxyType
from typing import List, Iterable, Union, Set, Optional, Type, Tuple, Mapping, Callable, Sequence
from collections import ChainMap

import numpy as np

import rdkit
import scm.plams.interfaces.molecule.rdkit as molkit
from rdkit.Chem import AllChem, Mol
from scm.plams.core.basejob import Job
from scm.plams import (Molecule, Atom, Bond, MoleculeError, add_to_class, Units,
                       Settings, AMSJob, ADFJob, Cp2kJob, axis_rotation_matrix)

from .mol_split_cm import SplitMol
from .remove_atoms_cm import RemoveAtoms
from .optimize_rotmat import optimize_rotmat
from ..logger import logger
from ..workflows import WorkFlow, MOL
from ..mol_utils import fix_carboxyl, to_atnum
from ..settings_dataframe import SettingsDataFrame
from ..data_handling.mol_to_file import mol_to_file
from ..jobs import job_geometry_opt  # noqa: F401
from ..utils import KindEnum, AnchorTup

__all__ = ['init_ligand_opt']

UFF = AllChem.UFFGetMoleculeForceField


def init_ligand_opt(ligand_df: SettingsDataFrame) -> None:
    """Initialize the ligand optimization procedure."""
    workflow = WorkFlow.from_template(ligand_df, name='ligand_opt')

    # Pull from the database; push unoptimized structures
    idx = workflow.from_db(ligand_df, get_mol=True)
    workflow.to_db(ligand_df, index=idx)
    if not ligand_df.settings.optional.ligand.optimize:
        return None

    # Start the ligand optimization
    workflow(start_ligand_jobs, ligand_df, columns=[], index=idx)

    # Push the optimized structures to the database
    job_recipe = workflow.get_recipe()
    workflow.to_db(ligand_df, status='optimized', index=idx, job_recipe=job_recipe)

    # Export ligands to .xyz, .pdb, .mol and/or .mol format
    mol_format = ligand_df.settings.optional.database.mol_format
    if mol_format:
        path = workflow.path
        mol_to_file(ligand_df.loc[idx, MOL], path, mol_format=mol_format)


def _set_charge_adfjob(s: Settings, charge: int) -> None:
    s.input.charge = charge


def _set_charge_amsjob(s: Settings, charge: int) -> None:
    if s.get('uff') is not None:
        return
    s.input.ams.system.charge = charge


def _set_charge_cp2kjob(s: Settings, charge: int) -> None:
    if s.get('dft') is None:
        return
    s.input.force_eval.dft.charge = charge


ChargeFunc = Callable[[Settings, int], None]
CHARGE_FUNC_MAPPING: Mapping[Type[Job], ChargeFunc] = MappingProxyType({
    ADFJob: _set_charge_adfjob,
    AMSJob: _set_charge_amsjob,
    Cp2kJob: _set_charge_cp2kjob,
})


def start_ligand_jobs(ligand_list: Iterable[Molecule],
                      jobs: Iterable[Optional[Type[Job]]],
                      settings: Iterable[Optional[Settings]],
                      use_ff: bool = False, **kwargs) -> None:
    """Loop over all molecules in ``ligand_df.loc[idx]`` and perform geometry optimizations."""
    _j1, job = jobs
    _s1, s = settings

    if job not in {None, AMSJob, ADFJob, Cp2kJob}:
        raise NotImplementedError(f"job1: {job.__class__} = {job!r}")
    elif use_ff:
        raise NotImplementedError(f"use_ff: {use_ff.__class__} = {use_ff!r}")

    if job is None:
        _start_ligand_jobs_uff(ligand_list)
    else:
        charge_func = CHARGE_FUNC_MAPPING[job]
        _start_ligand_jobs_plams(ligand_list, job, s, charge_func)
    return None


def _start_ligand_jobs_plams(ligand_list: Iterable[Molecule],
                             job: Type[Job], settings: Settings,
                             charge_func: ChargeFunc) -> None:
    """Loop over all molecules in ``ligand_df.loc[idx]`` and perform geometry optimizations."""
    for ligand in ligand_list:
        try:
            optimize_ligand(ligand)
        except Exception as ex:
            logger.debug(f'{ex.__class__.__name__}: {ex}', exc_info=True)
            ligand.properties.is_opt = False
            continue

        s = Settings(settings)
        charge_func(s, int(sum(at.properties.get('charge', 0) for at in ligand)))
        ligand.job_geometry_opt(job, s, name='ligand_opt')
        ligand.round_coords()  # `.is_opt = True` is set by `ligand.job_geometry_opt()``
    return None


def _start_ligand_jobs_uff(ligand_list: Iterable[Molecule]) -> None:
    """Loop over all molecules in ``ligand_df.loc[idx]`` and perform geometry optimizations."""
    for ligand in ligand_list:
        logger.info(f'UFFGetMoleculeForceField: {ligand.properties.name} optimization has started')
        try:
            optimize_ligand(ligand)
        except Exception as ex:
            ligand.properties.is_opt = False
            logger.error(f'UFFGetMoleculeForceField: {ligand.properties.name} optimization '
                         f'has failed: {ex!r}', exc_info=True)
            logger.debug(f'{ex.__class__.__name__}: {ex}', exc_info=True)
        else:
            ligand.properties.is_opt = True
            logger.info(f'UFFGetMoleculeForceField: {ligand.properties.name} optimization '
                        'is successful')
    return None


def optimize_ligand(ligand: Molecule) -> None:
    """Optimize a ligand molecule."""
    anchor = ligand.properties.dummies

    # Split the branched ligand into linear fragments and optimize them individually
    bonds = split_mol(ligand, anchor)
    context = SplitMol(ligand, bonds)
    with context as mol_frags:
        cap_dict = ChainMap(*context._at_pairs)
        for mol in mol_frags:
            cap_list = [cap for at, cap in cap_dict.items() if at in mol]
            mol.set_dihed(180.0, anchor, cap_list)

    # Find the optimal dihedrals angle between the fragments
    for bond in bonds:
        modified_minimum_scan_rdkit(ligand, ligand.get_index(bond))

    # RDKit UFF can sometimes mess up the geometries of carboxylates: fix them
    fix_carboxyl(ligand)

    # Allign the ligand with the Cartesian X-axis.
    allign_axis(ligand)


def _get_allign_args(
    mol: "Molecule | Mol",
    anchor_tup: AnchorTup,
) -> Tuple[np.ndarray, int, int, Tuple[int, int, int], "None | float"]:
    if anchor_tup.kind == KindEnum.FIRST:
        idx_rot = idx_trans = anchor_tup.anchor_idx[0]
        xyz = mol.as_array() if isinstance(mol, Molecule) else rdmol_as_array(mol)
        return xyz, idx_rot, idx_trans, anchor_tup.anchor_idx[:3], anchor_tup.angle_offset

    # Reserve index `-1` for a dummy atom representing the mean position of all anchors
    if isinstance(mol, Molecule):
        xyz = np.empty((len(mol) + 1, 3), dtype=np.float64)
        xyz[:-1] = mol
    else:
        xyz = np.empty((len(mol.GetAtoms()) + 1, 3), dtype=np.float64)
        xyz[:-1] = rdmol_as_array(mol)
    xyz[-1] = xyz[np.array(anchor_tup.anchor_idx)].mean(axis=0)

    if anchor_tup.kind == KindEnum.MEAN:
        idx_rot = idx_trans = -1
    elif anchor_tup.kind == KindEnum.MEAN_TRANSLATE:
        idx_rot = -1
        idx_trans = anchor_tup.anchor_idx[0]
    else:
        raise ValueError(f"Unknown anchor kind: {anchor_tup.kind!r}")
    return xyz, idx_rot, idx_trans, anchor_tup.anchor_idx[:3], anchor_tup.angle_offset


def allign_axis(mol: Molecule) -> None:
    """Allign a molecule with the Cartesian X-axis; setting **anchor** as the origin."""
    xyz, idx_rot, idx_trans, idx_angle, angle = _get_allign_args(mol, mol.properties.anchor_tup)
    rotmat1 = optimize_rotmat(xyz, idx_rot)
    xyz = np.matmul(xyz, rotmat1.T, out=xyz)

    # Manually rotate the ligand by a user-specified amount
    if angle is not None:
        i, j, k = idx_angle
        vec_perp = np.cross(xyz[j] - xyz[i], xyz[k] - xyz[j])
        rotmat2 = axis_rotation_matrix(vec_perp, angle)
        xyz = np.matmul(xyz, rotmat2.T, out=xyz)

    xyz -= xyz[idx_trans]
    mol.from_array(xyz.round(decimals=3))


def split_mol(plams_mol: Molecule, anchor: Atom) -> List[Bond]:
    """Split a molecule into multiple smaller fragments; returning the bonds that have to be broken.

    One fragment is created for every branch within **plams_mol**.

    Parameters
    ----------
    plams_mol : |plams.Molecule|
        The input molecule.

    anchor : |plams.Atom|
        An anchor atom which will be stored in the largest to-be returned fragment.

    Returns
    -------
    :class:`list` [|plams.Bond|]
        A list of plams bonds.

    """
    def _in_ring(bond: Bond) -> bool:
        """Check if **bond** is part of a ring system."""
        return plams_mol.in_ring(bond)

    def _besides_ring(atom: Atom) -> bool:
        """Check if any neighboring atoms are part of a ring system."""
        return any(plams_mol.in_ring(at) for at in atom.neighbors())

    def _is_valid_bond(bond: Bond) -> bool:
        """Check if one atom in **bond** has at least 3 neighbours and the other at least 2."""
        n1, n2 = plams_mol.neighbors(bond.atom1), plams_mol.neighbors(bond.atom2)
        return (len(n1) >= 3 and len(n2) >= 2) or (len(n1) >= 2 and len(n2) >= 3)

    def _get_frag_size(bond: Bond) -> int:
        """Return the size of the largest fragment were **plams_mol** to be split along **bond**."""
        if getattr(bond, '_besides_ring', False):
            del bond._besides_ring
            return np.inf
        return plams_mol.get_frag_size(bond, anchor)

    # Temporary remove hydrogen atoms
    atom_gen = (at for at in plams_mol if at.atnum == 1)
    with RemoveAtoms(plams_mol, atom_gen):
        # Remove undesired bonds
        bond_list = [bond for bond in plams_mol.bonds if not _in_ring(bond)]

        # Remove even more undesired bonds
        for bond in reversed(bond_list):
            if not _is_valid_bond(bond):
                bond_list.remove(bond)

    atom_list = list(itertools.chain.from_iterable((bond.atom1, bond.atom2) for bond in bond_list))
    atom_ge3 = (atom for atom in atom_list if atom_list.count(atom) >= 3)
    atom_dict = {atom: [bond for bond in atom.bonds if bond in bond_list] for atom in atom_ge3}
    for b in bond_list:
        if plams_mol.in_ring(b.atom1):
            key = b.atom1
        elif plams_mol.in_ring(b.atom2):
            key = b.atom2
        else:
            continue

        b._besides_ring = True
        if key not in atom_dict:
            atom_dict[key] = 3 * [b]
        else:
            atom_dict[key] += (3 - len(atom_dict[key])) * [b]

    # Fragment the molecule such that the anchor on the largest fragment
    ret = []
    for at, bond_list in atom_dict.items():
        # Can't directly iterate over bond_list as its size is modified
        iterator = range(len(bond_list[2:]))
        for _ in iterator:
            frag_size = [_get_frag_size(bond) for bond in bond_list]
            idx = np.argmax(frag_size)  # The index of the largest fragment
            bond = bond_list.pop(idx)
            ret.append(bond)
    return ret


@add_to_class(Molecule)
def get_frag_size(self, bond: Bond, anchor: Atom) -> int:
    """Return the size of the fragment containing **atom** if this instance was split into two by the breaking of **bond**.

    Parameters
    ----------
    bond : |plams.Bond|
        A PLAMS bond.

    anchor : |plams.Atom|
        A PLAMS atom. The size of the fragment containg this atom will be returned.

    Returns
    -------
    :class:`int`
        The number of atoms in the fragment containing **atom**.

    """  # noqa
    if bond not in self.bonds:
        raise MoleculeError('get_frag_size: The argument bond should be of type plams.Bond and '
                            'be part of the Molecule')
    elif anchor not in self.atoms:
        raise MoleculeError('get_frag_size: The argument atom should be of type plams.Atom and '
                            'be part of the Molecule')

    for at in self:
        at._visited = False

    frag1 = set()
    frag2 = set()

    def dfs(at1: Atom, atom_set: Set[Atom]):
        at1._visited = True
        atom_set.add(at1)
        for at2 in at1.neighbors():

            if at2._visited:
                continue
            dfs(at2, atom_set)

    bond.atom1._visited = bond.atom2._visited = True
    dfs(bond.atom1, frag1)
    dfs(bond.atom2, frag2)
    for at in self.atoms:
        del at._visited

    # fragment #1 contains **atom**
    if anchor in frag1:
        if bond.atom1 not in frag1:
            bond.atom1, bond.atom2 = bond.atom2, bond.atom1
        return len(frag1)

    if bond.atom1 not in frag2:
        bond.atom1, bond.atom2 = bond.atom2, bond.atom1
    return len(frag2)


def get_dihed(atoms: Iterable[Atom], unit: str = 'degree') -> float:
    """Return the dihedral angle defined by a set of four atoms.

    Parameters
    ----------
    atoms : |Iterable| [|plams.atoms|]
        An iterable consisting of 4 PLAMS atoms

    unit : :class:`str`
        The output unit.

    Returns
    -------
    :class:`float`
        A dihedral angle expressed in **unit**.

    """
    at1, at2, at3, at4 = atoms
    vec1 = -np.array(at1.vector_to(at2))
    vec2 = np.array(at2.vector_to(at3))
    vec3 = np.array(at3.vector_to(at4))

    v1v2, v2v3 = np.cross(vec1, vec2), np.cross(vec3, vec2)
    v1v2_v2v3 = np.cross(v1v2, v2v3)
    v2_norm_v2 = vec2 / np.linalg.norm(vec2)
    epsilon = np.arctan2(v1v2_v2v3@v2_norm_v2, v1v2@v2v3)

    return Units.convert(epsilon, 'radian', unit)


@add_to_class(Molecule)
def neighbors_mod(self, atom: Atom, exclude: Union[int, str] = 1) -> List[Atom]:
    """A modified PLAMS function: Allows the exlucison of specific elements from the return list.

    Return a list of neighbors of **atom** within the molecule.
    Atoms with **atom** has to belong to the molecule.
    Returned list follows the same order as the **atom.bond** attribute.

    Parameters
    ----------
    atom : |plams.Atom|_
        The plams atom whose neighbours will be returned.

    exclude : |str|_ or |int|_
        Exclude all neighbours with a specific atomic number or symbol.

    Returns
    -------
    |list|_ [|plams.Atom|_]
        A list of all neighbours of **atom**.

    """
    exclude = to_atnum(exclude)
    if atom.mol != self:
        raise MoleculeError('neighbors: passed atom should belong to the molecule')
    return [b.other_end(atom) for b in atom.bonds if b.other_end(atom).atnum != exclude]


@add_to_class(Molecule)
def set_dihed(self, angle: float, anchor: Atom, cap: Sequence[Atom],
              opt: bool = True, unit: str = 'degree') -> None:
    """Change all valid dihedral angles into a specific value.

    Performs an inplace update of this instance.

    Parameters
    ----------
    angle : :class:`float`
        The desired dihedral angle.

    anchor : |plams.Atom|
        The ligand anchor atom.

    opt : :class:`bool`
        Whether or not the dihedral adjustment should be followed up by an RDKit UFF optimization.

    unit : :class:`str`
        The input unit.

    """
    cap_atnum = []
    for at in cap:
        cap_atnum.append(at.atnum)
        at.atnum = 0

    angle = Units.convert(angle, unit, 'degree')
    bond_iter = (bond for bond in self.bonds if bond.atom1.atnum != 1 and bond.atom2.atnum != 1
                 and bond.order == 1 and not self.in_ring(bond))

    # Correction factor for, most importantly, tri-valent anchors (e.g. P(R)(R)R)
    dihed_cor = angle / 2
    neighbors = anchor.neighbors()
    if len(neighbors) > 2:
        atom_list = [anchor] + sorted(neighbors, key=lambda at: -at.atnum)[:3]
        improper = get_dihed(atom_list)
        dihed_cor *= np.sign(improper)

    for bond in bond_iter:
        # Gather lists of all non-hydrogen neighbors
        n1, n2 = self.neighbors_mod(bond.atom1), self.neighbors_mod(bond.atom2)

        # Remove all atoms in `bond`
        n1 = [atom for atom in n1 if atom is not bond.atom2]
        n2 = [atom for atom in n2 if atom is not bond.atom1]

        # Remove all non-subsituted atoms
        # A special case consists of anchor atoms; they can stay
        if len(n1) > 1:
            n1 = [atom for atom in n1 if
                  (len(self.neighbors_mod(atom)) > 1 or atom is anchor or atom.atnum == 0)]
        if len(n2) > 1:
            n2 = [atom for atom in n2 if
                  (len(self.neighbors_mod(atom)) > 1 or atom is anchor or atom.atnum == 0)]

        # Set `bond` in an anti-periplanar conformation
        if n1 and n2:
            dihed = get_dihed((n1[0], bond.atom1, bond.atom2, n2[0]))
            if anchor not in bond:
                self.rotate_bond(bond, bond.atom1, angle - dihed, unit='degree')
            else:
                dihed -= dihed_cor
                self.rotate_bond(bond, bond.atom1, -dihed, unit='degree')
                dihed_cor *= -1

    for at, atnum in zip(cap, cap_atnum):
        at.atnum = atnum

    if opt:
        rdmol = molkit.to_rdmol(self)
        UFF(rdmol).Minimize()
        self.from_rdmol(rdmol)


def rdmol_as_array(rdmol: rdkit.Chem.Mol) -> np.ndarray:
    """Convert an rdkit molecule into an array of Cartesian coordinates."""
    def get_xyz(atom: rdkit.Chem.Atom) -> Tuple[float, float, float]:
        pos = conf.GetAtomPosition(atom.GetIdx())
        return (pos.x, pos.y, pos.z)

    conf = rdmol.GetConformer(id=-1)
    atoms = rdmol.GetAtoms()

    atom_count = len(atoms)
    count = 3 * atom_count
    shape = atom_count, 3

    iterator = itertools.chain.from_iterable(get_xyz(at) for at in atoms)
    ret = np.fromiter(iterator, count=count, dtype=float)
    ret.shape = shape
    return ret


def _find_idx(mol: Molecule, bond: Bond) -> List[int]:
    """Return the atomic indices of all atoms on the side of **bond.atom2**."""
    ret = []
    mol.set_atoms_id(start=0)
    for at in mol:
        at._visited = False

    def dfs(at1, mol):
        at1._visited = True
        ret.append(at1.id)
        for bond in at1.bonds:
            at2 = bond.other_end(at1)
            if not at2._visited:
                dfs(at2, mol)

    bond.atom1._visited = bond.atom2._visited = True
    dfs(bond.atom2, mol)

    mol.unset_atoms_id()
    return ret


def modified_minimum_scan_rdkit(ligand: Molecule, bond_tuple: Tuple[int, int]) -> None:
    """A modified version of the :func:`.global_minimum_scan_rdkit` function.

    * Uses the ligand vector as criteria rather than the energy.
    * Geometry optimizations are constrained during the conformation search.
    * Finish with a final unconstrained geometry optimization.

    See Also
    --------
    :func:`global_minimum_scan_rdkit<scm.plams.recipes.global_minimum.minimum_scan_rdkit>`:
        Optimize the molecule (RDKit UFF) with 3 different values for the given dihedral angle and
        find the lowest energy conformer.

        :param |Molecule| mol: The input molecule
        :param tuple bond_tuple: A 2-tuples containing the atomic indices of valid bonds
        :return |Molecule|: A copy of *mol* with a newly optimized geometry

    """
    # Define a number of variables and create 3 copies of the ligand
    angles = (-120, 0, 120)
    mol_list = [ligand.copy() for _ in range(3)]
    for angle, mol in zip(angles, mol_list):
        bond = mol[bond_tuple]
        atom = mol[bond_tuple[0]]
        mol.rotate_bond(bond, atom, angle, unit='degree')
    rdmol_list = [molkit.to_rdmol(mol, properties=False) for mol in mol_list]

    # Optimize the (constrained) geometry for all dihedral angles in angle_list
    # The geometry that yields the minimum energy is returned
    fixed = _find_idx(mol, bond)
    for rdmol in rdmol_list:
        ff = UFF(rdmol)
        for f in fixed:
            ff.AddFixedPoint(f)
        ff.Minimize()

    # Find the conformation with the optimal ligand vector
    cost_list = []
    for rdmol in rdmol_list:
        xyz, idx_rot, idx_trans, idx_angle, angle = _get_allign_args(rdmol,
                                                                     ligand.properties.anchor_tup)
        rotmat = optimize_rotmat(xyz, idx_rot)
        xyz = np.matmul(xyz, rotmat.T, out=xyz)
        if angle is not None:
            i, j, k = idx_angle
            vec_perp = np.cross(xyz[j] - xyz[i], xyz[k] - xyz[j])
            rotmat2 = axis_rotation_matrix(vec_perp, angle)
            xyz = np.matmul(xyz, rotmat2.T, out=xyz)
        xyz -= xyz[idx_trans]
        cost = np.exp(xyz[:, 1:]).sum()
        cost_list.append(cost)

    # Perform an unconstrained optimization on the best geometry and update the geometry of ligand
    j = np.argmin(cost_list)
    rdmol_best = rdmol_list[j]
    UFF(rdmol).Minimize()
    ligand.from_rdmol(rdmol_best)



"""Substitution symmetry."""

import numpy as np
import pandas as pd
from scipy.spatial.distance import cdist

from scm.plams import Molecule
from scm.plams.tools.geometry import rotation_matrix

from ..logger import logger

__all__ = ['del_equiv_structures']


def find_equivalent_atoms(mol, idx=None, idx_substract=0):
    """Take a molecule, 'mol', and return the indices of all symmetry equivalent atoms.

    The implemented function is based on finding duplicates in the (sorted) distance matrix,
    as symmetry equivalent atoms have identical inter-atomic distances.

    Parameters
    ----------
    mol : |plams.Molecule| or numpy array
        A PLAMS molecule or numpy array
    idx : None, int or list
        An iterable consisting of atomic indices. ALl atoms in 'mol' will be examined if None
    idx_substract : int
        Substract a constant from all values in 'idx';
        usefull for interconverting between 1-based and 0-based indices

    Returns
    -------
    list
        A list with tuples of symmetry equivalent atomic indices.

    """
    # Convert a PLAMS molecule to an array
    if isinstance(mol, Molecule):
        mol = mol.as_array()

    # If **idx** is *None*, investigate all atoms in **mol**
    if idx is None:
        idx = slice(0, len(mol))
    else:
        idx = np.array(idx, dtype=int) - idx_substract

    # Create a distance matrix, round it to 2 decimals and isolate all unique rows
    dist = np.around(cdist(mol, mol), decimals=2)
    dist.sort(axis=1)
    unique_at = np.unique(dist[idx], axis=0)

    # Find and return the indices of all duplicate rows in the distance matrix
    return [tuple(j for j, ar2 in enumerate(dist) if (ar1 == ar2).all()) for ar1 in unique_at]


def reset_origin(mol, at1):
    """Reset the origin of a molecule by means of translations and rotations.

    Parameters
    ----------
    mol : |plams.Molecule| or np.ndarray
        A PLAMS molecule or array of xyz coordinates.
    idx : int
        The atomic index of the atom that will be alligned to the x-axis.

    Returns
    -------
    |plams.Molecule|
        The rotated and translated PLAMS molecule

    """
    if isinstance(mol, Molecule):
        mol = mol.as_array()

    # Translate
    mol -= mol.mean(axis=0)

    # Rotate
    vec1 = mol[at1] - np.zeros(3)
    vec2 = np.array([1, 0, 0])
    rotmat = rotation_matrix(vec1, vec2)
    mol = mol@rotmat

    return mol


def get_rotmat_axis(rot_range, axis='x'):
    """Calculate the rotation matrix for rotating a vector along an axis.

    A rotation matrix is constructed for each angle in **rot_range**.

    Parameters
    ----------
    rot_range : |np.ndarray|
        An array of rotations in radian.

    """
    ret = np.zeros((len(rot_range), 3, 3))
    ret[:, 0, 0] = ret[:, 1, 1] = ret[:, 2, 2] = np.ones(len(rot_range))

    if axis.lower() == 'x':  # Rotation around x axis
        ret[:, 1, 1] = ret[:, 2, 2] = np.cos(rot_range)
        ret[:, 2, 1] = ret[:, 1, 2] = np.sin(rot_range)
        ret[:, 2, 1] *= -1
    elif axis.lower() == 'y':  # Rotation around y axis
        ret[:, 0, 0] = ret[:, 2, 2] = np.cos(rot_range)
        ret[:, 2, 0] = ret[:, 0, 2] = np.sin(rot_range)
        ret[:, 2, 0] *= -1
    elif axis.lower() == 'z':  # Rotation around z axis
        ret[:, 0, 0] = ret[:, 1, 1] = np.cos(rot_range)
        ret[:, 0, 1] = ret[:, 1, 0] = np.sin(rot_range)
        ret[:, 0, 1] *= -1

    return ret


def supstitution_symmetry(mol):
    """Returns atomic symbols of substituted atoms (or first conection of non diatomic ligand).

    Writes type of substitution symetry at the molecular properties

    Parameters
    ----------
    mol : |plams.Molecule|
        A PLAMS molecule

    Returns
    -------
    str
        Type of subsymmetry

    """
    dataframe, type_of_symetry = [], []
    ligand_identity = mol.properties.ligID

    # Defining C atoms conected to substituents and making Molecule object (cmol) out of them
    catoms = mol.properties.coords_other_arrays
    cmol = Molecule()
    for at in catoms:
        cmol.add_atom(mol.closest_atom(at))

    # If substitution is linear, simple combinations without repetition can be applied
    if len(ligand_identity) <= 2:
        if len(ligand_identity) == 1:
            logger.warning("One does not simply ask for subsymmetry of one atom!")
            return
        elif len(ligand_identity) == 0:
            logger.warning("One does not simply ask for subsymmetry of no atom")
            return
        else:
            subsymmetry = 'linear'
    else:
        # Getting non zero row indices from data frame - defines symmetry type

        dataframe = get_symmetry(cmol, decimals=2)
        type_of_symetry = np.unique(dataframe.to_numpy().nonzero()[0])

        # Assign type of symetry and atomic symbols
        if list(type_of_symetry) == [0, 1, 8, 9]:
            subsymmetry = 'D2h'
        else:
            logger.warning("Subsymmetry is not recognized")
            return

    return subsymmetry


def get_symmetry(mol, decimals=2):
    """Return the number of equivalent atoms under a number of symmetry operations.

    Parameters
    ----------
    mol : |plams.Molecule|
       A PLAMS molecule or array of xyz coordinates.
    decimals  : int
        The error marigin (number of decimals) in determining equivalent atoms

    Returns
    -------
    |pd.DataFrame|
        A Pandas dataframe with the number of equivalent atoms per axis per operation.

    """
    if isinstance(mol, Molecule):
        mol = mol.as_array()

    # Prepare the dataframe
    columns = ['x', 'y', 'z']
    index = ['2pi / ' + str(i) for i in range(1, 9)] + ['reflection', 'inversion']
    data = np.empty((len(index), 3))
    df = pd.DataFrame(data=data, index=index, columns=columns)
    df.index.name = 'Rotations, reflections and inversions'
    df.columns.name = 'Identical atoms'

    # Fill the dataframe
    rot_range = (2 * np.pi) / np.arange(1, 9)
    for i, j in enumerate(columns):
        # Rotate
        rotmat = get_rotmat_axis(rot_range, axis=j)
        mol_new = np.empty((10, mol.shape[0], 3))
        mol_new[0:8] = mol@rotmat

        # Mirror
        mol_new[8] = mol.copy()
        mol_new[8][i] *= -1

        # Invert
        mol_new[9] = -1 * mol.copy()

        # Calculate the number of equivalent atoms
        dist_mat = np.array([cdist(i, mol) for i in mol_new]).round(decimals)
        try:
            df[j] = np.bincount(np.where(dist_mat == 0)[0])
        except ValueError:
            logger.warning("Something went wrong: Length of values does not match length of index")
    return df


def del_equiv_structures(mols, subsymmetry=None):
    """Returne a list of unique molecules based on subsymmetry.

    Permutes list of ligands form plams_mol.properties.ligID for each molecule
    Molecules that have identical list of permutations are equivalent

    Parameters
    ----------
    mol : |plams.Molecule|
       A PLAMS molecule or array of xyz coordinates.
    subsymmetry : str
        A type of subsymmetry

    Returns
    -------
    list
        A list of unique molecule for specific subsymmetry

    """
    notunique = []

    for mol in mols:
        if subsymmetry is None:
            subsymmetry = supstitution_symmetry(mol)

        lig_id = list(mol.properties.ligID)

        all_permutations = symm_permutations(subsymmetry, lig_id)

        notunique.append(all_permutations)

    try:
        scos = [sorted(sc) for sc in notunique]
    except TypeError:
        # notunique == [None, ...] if something went wrong in the previous steps
        # assume that all molecules are unique in such a scenario
        return mols

    u, indices = np.unique(scos, return_index=True, axis=0)

    unique_molecules = [mols[i] for i in list(indices)]

    return unique_molecules


def symm_permutations(condition, elements):
    """For given list of elements, makes permutations taking in account symmetry condition.

    Parameters
    ----------
    condition : string
        Type of subsymmetry
    elements : list
        A list of integers to be permuted

    """

    def swap_neighbours(j):
        """Swap neighbours inside a list: 1, 2, 3, 4 becomes 2, 1, 4, 3.

        <j>: a list
        """
        j[1::2], j[::2] = j[::2], j[1::2]
        return j

    def swap_pairs(j):
        """Swap pairs inside a list: 1,2,3,4 becomes 3,4,1,2.

        <j>: a list
        """
        j[:2], j[2:] = j[2:], j[:2]
        return j

    def rotate_list(lst, n):
        return lst[n:] + lst[:n]

    def swap_two_last(j):
        j[-1], j[-2] = j[-2], j[-1]
        return j

    # Making list of all permutations
    if condition is None:
        return None
    if condition == 'D2h':
        a = ''.join(elements)
        b = ''.join(swap_neighbours(elements))
        c = ''.join(swap_pairs(elements))
        d = ''.join(swap_neighbours(elements))
        final = []
        final = [a, b, c, d]
    if condition == 'linear':
        a = ''.join(elements)
        b = ''.join(swap_neighbours(elements))
        final = []
        final = [a, b]
    if condition == 'triangle':
        a = ''.join(elements)
        b = ''.join(swap_two_last(elements))
        final = []
        final = [a, b]
    return final



"""A module designed for sanitizing and interpreting the input file.

Index
-----
.. currentmodule:: CAT.data_handling.validation_schemas
.. autosummary::
    mol_schema
    core_schema
    ligand_schema
    qd_schema
    database_schema
    mongodb_schema
    ligand_opt_schema
    bde_schema
    qd_opt_schema
    crs_schema
    asa_schema
    subset_schema
    multi_ligand_schema
    JOB_MAP
    JOB_MAP_SCM

API
---
.. autodata:: mol_schema
    :annotation: : schema.Schema
.. autodata:: core_schema
    :annotation: : schema.Schema
.. autodata:: ligand_schema
    :annotation: : schema.Schema
.. autodata:: qd_schema
    :annotation: : schema.Schema
.. autodata:: database_schema
    :annotation: : schema.Schema
.. autodata:: mongodb_schema
    :annotation: : schema.Schema
.. autodata:: ligand_opt_schema
    :annotation: : schema.Schema
.. autodata:: bde_schema
    :annotation: : schema.Schema
.. autodata:: qd_opt_schema
    :annotation: : schema.Schema
.. autodata:: asa_schema
    :annotation: : schema.Schema
.. autodata:: crs_schema
    :annotation: : schema.Schema
.. autodata:: subset_schema
    :annotation: : schema.Schema
.. autodata:: multi_ligand_schema
    :annotation: : schema.Schema
.. autodata:: JOB_MAP
    :annotation: : Mapping[str, Type[Job]]
.. autodata:: JOB_MAP_SCM
    :annotation: : Mapping[str, Type[Job]]

"""

from types import MappingProxyType
from typing import Collection, Callable, Any, Optional, TypeVar, Mapping, Type, Tuple
from operator import __index__
from collections import abc

import numpy as np
from schema import Or, And, Use, Schema
from schema import Optional as Optional_

from scm.plams import CRSJob, Settings
from scm.plams.core.basejob import Job

from scm.plams.interfaces.adfsuite.adf import ADFJob
from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.interfaces.adfsuite.uff import UFFJob
from scm.plams.interfaces.adfsuite.band import BANDJob
from scm.plams.interfaces.adfsuite.dftb import DFTBJob
from scm.plams.interfaces.adfsuite.mopac import MOPACJob
from scm.plams.interfaces.adfsuite.reaxff import ReaxFFJob

from scm.plams.interfaces.thirdparty.cp2k import Cp2kJob
from scm.plams.interfaces.thirdparty.orca import ORCAJob
from scm.plams.interfaces.thirdparty.dirac import DiracJob
from scm.plams.interfaces.thirdparty.gamess import GamessJob
from scm.plams.interfaces.thirdparty.dftbplus import DFTBPlusJob

try:
    from dataCAT import Database
except ImportError:
    Database = None

from .str_to_func import str_to_func
from ..utils import get_template, validate_path, validate_core_atom, check_sys_var
from ..mol_utils import to_atnum
from ..utils import AllignmentEnum, AllignmentTup

__all__ = ['mol_schema', 'core_schema', 'ligand_schema', 'qd_schema', 'database_schema',
           'mongodb_schema', 'bde_schema', 'ligand_opt_schema', 'qd_opt_schema', 'crs_schema',
           'asa_schema', 'subset_schema', 'bulkiness_schema']


def val_float(value: Any) -> bool:
    """Check if a float-like object has been passed (:data:`typing.SupportsFloat`)."""
    try:
        value.__float__()
        return True
    except Exception:
        return False


def val_int(value: Any) -> bool:
    """Check if a int-like object has been passed (:data:`typing.SupportsInt`)."""
    try:
        value.__int__()
        return float(value).is_integer()
    except Exception:
        return False


def val_index(value: Any) -> bool:
    """Check if a index-like object has been passed (:data:`typing.SupportsIndex`)."""
    try:
        value.__index__()
        return True
    except Exception:
        return False


#: A dictionary for translating strings into :class:`plams.Job` types for third party software
JOB_MAP: Mapping[str, Type[Job]] = MappingProxyType({
    'cp2k': Cp2kJob, 'cp2kjob': Cp2kJob,
    'orca': ORCAJob, 'orcajob': ORCAJob,
    'dirac': DiracJob, 'diracjob': DiracJob,
    'gamess': GamessJob, 'gamessjob': GamessJob,
    'dftbplus': DFTBPlusJob, 'dftbplusjob': DFTBPlusJob,
})


#: A dictionary for translating strings into :class:`plams.Job` types for ADF
JOB_MAP_SCM: Mapping[str, Type[Job]] = MappingProxyType({
    'adf': ADFJob, 'adfjob': ADFJob,
    'ams': AMSJob, 'amsjob': AMSJob,
    'uff': UFFJob, 'uffjob': UFFJob,
    'band': BANDJob, 'bandjob': BANDJob,
    'dftb': DFTBJob, 'dftbjob': DFTBJob,
    'mopac': MOPACJob, 'mopacjob': MOPACJob,
    'reaxff': ReaxFFJob, 'reaxffjob': ReaxFFJob,
    'crs': CRSJob, 'cosmo-rs': CRSJob, 'crsjob': CRSJob
})


DB_NAMES: Tuple[str, ...] = ('core', 'ligand', 'qd')
FORMAT_NAMES: Tuple[str, ...] = ('pdb', 'xyz')
FORMAT_NAMES2: Tuple[str, ...] = ('pdb', 'xyz', 'mol', 'mol2')


def val_job_type(value: type) -> type:
    """Call :func:`.check_sys_var` if value is in :data:`.JOB_MAP_SCM`."""
    if value in JOB_MAP_SCM.values():
        check_sys_var()
    return value


def str_to_job_type(key: str) -> type:
    """Convert a string into a type object.

    Parameters
    ----------
    key : str
        An alias for a :class:`type` object of a PLAMS :class:`Job`.
        **key** is converted by passing it into :data:`.JOB_MAP` or :data:`.JOB_MAP_SCM`.
        The :func:`check_sys_var` function is called if **key** is present
        in :data:`.JOB_MAP_SCM`.

    Returns
    -------
    |type|_
        A type object constructed from **key**.

    Raises
    ------
    KeyError
        Raised if the decapitalized **key** is available in neither :data:`.JOB_MAP` nor
        :data:`.JOB_MAP_SCM`.

    """
    _key = key.lower()
    if _key in JOB_MAP:
        return JOB_MAP[_key]
    elif _key in JOB_MAP_SCM:
        check_sys_var()
        return JOB_MAP_SCM[_key]
    raise KeyError(f'No Job type alias available for {_key!r}')


T = TypeVar('T')


def to_tuple(collection: Collection[T], func: Optional[Callable[[T], Any]] = None) -> tuple:
    """Convert a collection to a sorted tuple."""
    try:
        ret = sorted(collection)
    except TypeError:  # The collection contains a mix of sorting-incompatible objects
        ret = sorted(collection, key=str)
    finally:
        if func is None:
            return tuple(ret)
        else:
            return tuple(func(i) for i in ret)


# The **default** parameter of schema.Optional() will automatically call any callable
# Solution: provide a callable that returns another callable
def _get_amsjob() -> type:
    """Return a type object of :class:`.AMSJob`."""
    return AMSJob


def _get_crsjob() -> type:
    """Return a type object of :class:`.CRSJob`."""
    return CRSJob


# Default settings templates
_bde_s1_default = get_template('qd.yaml')['MOPAC']
_bde_s2_default = get_template('qd.yaml')['UFF']

_qd_opt_s1_default = get_template('qd.yaml')['UFF']
_qd_opt_s2_default = _qd_opt_s1_default

_crs_s1_default = get_template('qd.yaml')['COSMO-MOPAC']
_crs_s2_default = get_template('qd.yaml')['COSMO-RS activity coefficient']
_crs_s2_default.update(get_template('crs.yaml')['MOPAC PM6'])


#: Schema for validating the ``['input_ligands']`` and ``['input_cores']`` blocks.
mol_schema: Schema = Schema({
    Optional_('guess_bonds', default=False):
        And(bool, error=".guess_bonds expects a boolean"),

    Optional_('is_core'):
        And(bool, error=".is_core expects a boolean"),

    Optional_('is_qd'):
        And(bool, error=".is_qd expects a boolean"),

    Optional_('ligand_smiles'):
        And(str, error=".ligand_smiles expects a string"),

    Optional_('ligand_anchor'):
        And(str, error=".ligand_smiles expects a string"),

    Optional_('column'):
        And(val_index, lambda n: n.__index__() >= 0, Use(int),
            error=".column expects an integer larger than or equal to 0"),

    Optional_('row'):
        And(val_index, lambda n: n.__index__() >= 0, Use(int),
            error=".row expects an integer larger than or equal to 0"),

    Optional_('parsed'): bool,

    Optional_('indices'):
        Or(
            And(
                val_index, lambda n: n.__index__() >= 0,
                Use(lambda n: to_tuple([n], func=__index__)),
                error=".indices expects an integer larger than or equal to 0"
            ),
            And(
                abc.Collection,
                lambda n: all(val_index(i) for i in n),
                lambda n: all(i.__index__() >= 0 for i in n),
                lambda n: len(n) == len(set(n)),
                Use(lambda n: to_tuple(n, func=__index__)),
                error=".indices expects one or more unique integers larger than or equal to 0"
            ),
            error=".indices expects an atomic index (int) or a list unique atomic indices"
        ),

    Optional_('type'):
        And(str, error='.type expects a string'),

    Optional_('name'):
        And(str, error='.name expects a string'),

    Optional_('path'):
        And(
            Use(validate_path),
            error=".path expects a None or a string pointing to an existing directory"
        ),
})

allignment_mapping = {
    'sphere': AllignmentTup(AllignmentEnum.SPHERE, False),
    'surface': AllignmentTup(AllignmentEnum.SURFACE, False),
    'sphere_invert': AllignmentTup(AllignmentEnum.SPHERE, True),
    'sphere invert': AllignmentTup(AllignmentEnum.SPHERE, True),
    'surface_invert': AllignmentTup(AllignmentEnum.SURFACE, True),
    'surface invert': AllignmentTup(AllignmentEnum.SURFACE, True),
}

#: Schema for validating the ``['optional']['core']`` block.
core_schema: Schema = Schema({
    'dirname':
        And(str, error='optional.core.dirname expects a string'),

    # Alias for `optional.core.anchor`
    Optional_('dummy', default=None): object,

    Optional_('anchor', default=None): object,

    Optional_('subset', default=None):
        Or(None, dict, error="optional.core.subset epected 'None' or a dictionary"),

    Optional_('allignment', default=AllignmentTup(AllignmentEnum.SURFACE, False)):
        Or(
            AllignmentTup,
            And(str, Use(lambda n: allignment_mapping[n.lower()]),
                error="optional.core.allignment expected "
                      f"one of {list(allignment_mapping.keys())}"),
        ),
})

#: Schema for validating the ``['optional']['core']['subset']`` block.
subset_schema: Schema = Schema({
    'f':
        And(val_float, lambda n: 0 < float(n) <= 1, Use(float)),

    Optional_('p'):
        And(val_float, lambda n: abs(float(n)) > 0, Use(float)),

    Optional_('mode', default='uniform'):
        And(str, lambda n: n.lower() in {'uniform', 'random', 'cluster'}, Use(str.lower)),

    Optional_('start', default=None):
        Or(
            None,
            And(val_index, Use(__index__))
        ),

    Optional_('follow_edge', default=False):
        bool,

    Optional_('cluster_size', default=1):
        Or(
            And(val_int, lambda n: int(n) > 0, Use(int)),
            And(abc.Collection,
                lambda n: all(val_int(i) for i in n),
                lambda n: all(int(i) > 0 for i in n),
                Use(lambda n: to_tuple(n, func=int)))
        ),

    Optional_('weight', default=lambda: str_to_func('np.exp(-x)')):
        Or(
            And(Callable, Use(str_to_func)),
            And(str, lambda n: 'x' in n, Use(str_to_func))
        ),

    Optional_('randomness', default=None):
        Or(
            None,
            And(val_float, lambda n: 0 <= float(n) <= 1, Use(float))
        )

})

#: Schema for validating the ``['optional']['database']`` block.
database_schema: Schema = Schema({
    # path+directory name of the database
    'dirname':
        And(str, error='optional.database.dirname expects a string'),

    Optional_('thread_safe', default=False): bool,

    Optional_('read', default=DB_NAMES):  # Attempt to pull structures from the database
        Or(
            And(None, Use(lambda n: ())),
            And(bool, Use(lambda n: DB_NAMES if n is True else ())),
            And(
                str,
                lambda n: n in DB_NAMES,
                Use(lambda n: (n,)),
                error=f'allowed values for optional.database.read are: {DB_NAMES!r}'
            ),
            And(
                abc.Collection,
                lambda n: all(i in DB_NAMES for i in n),
                lambda n: len(n) == len(set(n)),
                Use(to_tuple),
                error=f'allowed values for optional.database.read are: {DB_NAMES!r}'
            ),
            error='optional.database.read expects a boolean, string or list of unique strings'
        ),

    Optional_('write', default=DB_NAMES):  # Attempt to write structures to the database
        Or(
            And(None, Use(lambda n: ())),
            And(bool, Use(lambda n: DB_NAMES if n is True else ())),
            And(
                str,
                lambda n: n in DB_NAMES,
                Use(lambda n: (n,)),
                error=f'allowed values for optional.database.write are: {DB_NAMES!r}'
            ),
            And(
                abc.Collection,
                lambda n: all(i in DB_NAMES for i in n),
                lambda n: len(n) == len(set(n)),
                Use(to_tuple),
                error=f'allowed values for optional.database.write are: {DB_NAMES!r}'
            ),
            error='optional.database.write expects a boolean, string or list of unique strings'
        ),

    Optional_('overwrite', default=tuple):  # Allow previous entries to be overwritten
        Or(
            And(None, Use(lambda n: ())),
            And(bool, Use(lambda n: DB_NAMES if n is True else ())),
            And(
                str,
                lambda n: n in DB_NAMES,
                Use(lambda n: (n,)),
                error=f'allowed values for optional.database.overwrite are: {DB_NAMES!r}'
            ),
            And(
                abc.Collection,
                lambda n: all(i in DB_NAMES for i in n),
                lambda n: len(n) == len(set(n)),
                Use(to_tuple),
                error=f'allowed values for optional.database.overwrite are: {DB_NAMES!r}'
            ),
            error='optional.database.overwrite expects a boolean, string or list of unique strings'
        ),

    Optional_('mongodb', default=dict):  # Settings specific to MongoDB
        Or(
            dict,
            And(bool, lambda n: n is False, Use(lambda n: {})),
            error='optional.database.mongodb expects True (boolean) or a dictionary'
        ),

    Optional_('mol_format', default=FORMAT_NAMES):  # Return a tuple of file formats
        Or(
            And(None, Use(lambda n: ())),
            And(bool, Use(lambda n: FORMAT_NAMES if n is True else ())),
            And(
                str,
                lambda n: n in FORMAT_NAMES,
                error=f'allowed values for optional.database.mol_format are: {FORMAT_NAMES2!r}'
            ),
            And(
                abc.Collection,
                lambda n: all(i in FORMAT_NAMES for i in n),
                lambda n: len(n) == len(set(n)),
                Use(to_tuple),
                error=f'allowed values for optional.database.mol_format are: {FORMAT_NAMES2!r}'
            ),
            error='optional.database.mol_format expects a boolean, string or list of unique strings'
        ),

    Optional_('db'):
        Or(
            Database, None,
            error='optional.database.db expects None or a dataCAT.Database instance'
        ),
})


#: Schema for validating the ``['optional']['ligand']`` block.
ligand_schema: Schema = Schema({
    # path+directory name of the ligand directory
    'dirname':
        And(str, error='optional.ligand.dirname expects a string'),

    Optional_('anchor', default=None): object,

    # Alias for `optional.ligand.anchor`
    Optional_('functional_groups', default=None): object,

    Optional_('optimize', default={'job1': None}):  # Optimize the ligands
        Or(
            And(
                bool, Use(lambda n: ({'job1': None} if n else False)),
                error='optional.ligand.cosmo-rs expects a boolean or dictionary'
            ),
            And(
                dict,
                error='optional.ligand.cosmo-rs expects a boolean or dictionary'
            )
        ),

    Optional_('split', default=True):  # Remove a counterion from the function group
        And(bool, error='optional.ligand.split expects a boolean'),

    Optional_('cosmo-rs', default=False):  # Settings specific to ligand COSMO-RS calculations
        Or(
            dict,
            And(bool, Use(lambda n: {'job1': 'AMSJob'} if n else False)),
            error='optional.ligand.cosmo-rs expects a boolean or dictionary'
        ),

    Optional_('crs'):  # Settings specific to ligand COSMO-RS calculations
        Or(
            dict,
            And(bool, Use(lambda n: {'job1': 'AMSJob'} if n else False)),
            error='optional.ligand.cosmo-rs expects a boolean or dictionary'
        ),

    Optional_('cdft', default=False):  # Settings specific to ligand conceptual dft calculations
        Or(
            dict,
            And(bool, Use(lambda n: {'job1': 'ADFJob'} if n else False)),
            error='optional.ligand.cdft expects a boolean or dictionary'
        ),
})


#: Schema for validating the ``['optional']['qd']`` block.
qd_schema: Schema = Schema({
    # path+directory name of the quantum dot directory
    'dirname':
        And(str, error='optional.qd.dirname expects a string'),

    Optional_('construct_qd', default=True):  # Construct quantum dots
        Or(
            bool,
            error='optional.qd.construct_qd expects a boolean'
        ),

    # Settings specific to a quantum dot activation strain analyses
    Optional_('activation_strain', default=False):
        Or(
            And(bool, Use(lambda n: {'job1': None} if n else False)),
            dict, error='optional.qd.activation_strain expects a boolean or dictionary'
        ),

    Optional_('bulkiness', default=False):  # Ligand bulkiness workflow
        Or(
            And(bool, Use(lambda n: {'h_lim': 10.0, 'd': 'auto'} if n else False)),
            dict,
            error='optional.qd.bulkiness expects a boolean or dictionary'
        ),

    Optional_('optimize', default=False):  # Settings for quantum dot geometry optimizations
        Or(
            dict,
            And(bool, Use(lambda n: ({'job1': 'AMSJob'} if n else False))),
            error='optional.qd.optimize expects a boolean or dictionary'
        ),

    # Settings for quantum dot ligand dissociation calculations
    Optional_('dissociate', default=False):
        Or(
            dict,
            And(bool, lambda n: n is False),
            error='optional.qd.dissociate expects False (boolean) or a dictionary'
        ),

    Optional_('multi_ligand', default=None):  # Settings for multi-ligand attachment
        Or(
            None, dict,
            error='optional.qd.multi_ligand expects None or dictionary'
        ),
})

#: Schema for validating the ``['optional']['qd']['bulkiness']`` block.
bulkiness_schema = Schema({
    Optional_('h_lim', default=10.0): Or(
        And(val_float, lambda n: float(n) > 0, Use(float)),
        None,
        error='optional.qd.bulkiness.h_lim expects a positive float or None'
    ),

    Optional_('d', default='auto'): Or(
        And(val_float, lambda n: float(n) > 0, Use(float)),
        None,
        And(str, lambda n: n.lower() == 'auto', Use(str.lower)),
        error='optional.qd.bulkiness.d expects a positive float, None or "auto"'
    ),
})


#: Schema for validating the ``['optional']['database']['mongodb']`` block.
mongodb_schema: Schema = Schema({
    # Optional username for the MongoDB host
    Optional_('username'):
        Or(str, And(val_int, Use(int)),
           error='optional.database.mongodb.username expects a string or integer'),

    Optional_('password'):  # Optional password for the MongoDB host
        Or(str, And(val_int, Use(int)),
           error='optional.database.mongodb.password expects a string or integer'),

    Optional_('host', default='localhost'):  # Name of the MongoDB host
        Or(str, And(val_int, Use(int)),
           error='optional.database.mongodb.host expects a string or integer'),

    Optional_('port', default=27017):  # Port of the MongoDB host
        And(val_int, Use(int), error='optional.database.mongodb.port expects an integer'),

    Optional_(str):  # Other keyword arguments for :class:`pymongo.MongoClient`
        object
})


#: Schema for validating the ``['optional']['ligand']['optimize']`` block.
ligand_opt_schema: Schema = Schema({
    Optional_('use_ff', default=False):
        bool,

    # Delete files after the calculations are finished
    Optional_('keep_files', default=True):
        And(bool, error='optional.ligand.optimize.keep_files expects a boolean'),

    # The Job type and settings for the conformation search
    Optional_('job1', default=None): None,
    Optional_('s1', default=None): Or(None, dict),

    # The Job type for the final geometry optimization
    Optional_('job2', default=None):
        Or(
            None,
            And(
                And(type, lambda n: issubclass(n, Job), Use(val_job_type)),
                error=('optional.ligand.optimize.job1 expects a type object '
                       'that is a subclass of plams.Job')
            ),
            And(
                str, Use(str_to_job_type),
                error=('optional.ligand.optimize.job1 expects a string '
                       'that is a valid plams.Job alias')
            ),
            error='optional.ligand.optimize.job2 expects a string or a type object'
        ),

    # The Job Settings for the final geometry optimization
    Optional_('s2', default=Settings):
        Or(
            None,
            dict,
            And(str, Use(lambda n: get_template(n, from_cat_data=False))),
            error='optional.ligand.optimize.s2 expects a string or a dictionary'
        ),
})


#: Schema for validating the ``['optional']['qd']['dissociate']`` block.
bde_schema: Schema = Schema({
    Optional_('use_ff', default=False):
        bool,

    # Atom type of the to-be dissociated core atom
    Optional_('core_atom', default=None):
        Or(
            None,
            And(Or(And(val_int, Use(int)), str), Use(validate_core_atom)),
            error=('optional.qd.dissociate.core_atom expects a SMILES string, '
                   'atomic number (int) or atomic symbol (str)')
        ),

    Optional_('lig_count', default=None):  # The number of ligands per core_atom
        Or(
            None,
            And(val_int, lambda n: int(n) >= 0, Use(int)),
            error='optional.qd.dissociate.lig_count expects an integer larger than or equal to 0',
        ),

    Optional_('keep_files', default=True):  # Delete files after the calculations are finished
        And(bool, error='optional.qd.dissociate.keep_files expects a boolean'),

    Optional_('core_core_dist', default=None):
        Or(
            None,
            And(
                val_float, lambda n: float(n) > 0.0, Use(float),
                error=('optional.qd.dissociate.core_core_dist expects an integer or float '
                       'larger than or equal to 0.0')
            )
        ),

    Optional_('lig_core_dist', default=None):
        Or(
            None,
            And(
                val_float, lambda n: float(n) > 0.0, Use(float),
                error=('optional.qd.dissociate.lig_core_dist expects an integer or float '
                       'larger than or equal to 0.0')
            )
        ),

    Optional_('lig_core_pairs', default=1):
        Or(
            None,
            And(
                val_int, lambda n: int(n) > 0, Use(int),
                error=('optional.qd.dissociate.lig_pairs expects an integer'
                       'larger than or equal to 0')
            )
        ),

    Optional_('core_index', default=None):
        Or(
            None,
            And(
                val_index, lambda n: n.__index__() >= 0,
                Use(lambda n: to_tuple([n], func=__index__)),
                error=('optional.qd.dissociate.core_index expects an integer '
                       'larger than or equal to 0')
            ),
            And(
                abc.Collection,
                lambda n: all(val_index(i) for i in n),
                lambda n: all(i.__index__() >= 0 for i in n),
                lambda n: len(n) == len(set(n)),
                Use(lambda n: to_tuple(n, func=__index__)),
                error=('optional.qd.dissociate.core_index expects a list of unique integers '
                       'larger than or equal to 0')
            ),
            error=('optional.qd.dissociate.core_index expects an integer or list of unique integers'
                   'larger than or equal to 0')
        ),

    Optional_('topology', default=None):
        Or(
            None,
            And(
                abc.Mapping, lambda n: all(val_int(k) for k in n.keys()),
                Use(lambda n: {int(k): v for k, v in n.items()}),
                error='optional.qd.dissociate.topology expects a dictionary with integers as keys'
                )
        ),


    Optional_('job1', default=_get_amsjob):
        Or(
            And(
                And(type, lambda n: issubclass(n, Job), Use(val_job_type)),
                error=('optional.qd.dissociate.job1 expects a type object '
                       'that is a subclass of plams.Job')
            ),
            And(
                str, Use(str_to_job_type),
                error='optional.qd.dissociate.job1 expects a string that is a valid plams.Job alias'
            ),
            error='optional.qd.dissociate.job1 expects a string or a type object'

        ),

    Optional_('s1', default=_bde_s1_default.copy()):
        Or(
            dict,
            And(str, Use(lambda n: get_template(n, from_cat_data=False))),
            error='optional.qd.dissociate.s1 expects a string or a dictionary'
        ),

    Optional_('job2', default=None):
        Or(
            None,
            And(
                And(type, lambda n: issubclass(n, Job), Use(val_job_type)),
                error=('optional.qd.dissociate.job2 expects a type object '
                       'that is a subclass of plams.Job')
            ),
            And(
                str, Use(str_to_job_type),
                error='optional.qd.dissociate.job2 expects a string that is a valid plams.Job alias'
            ),
            error='optional.qd.dissociate.job2 expects a string or a type object'
        ),

    Optional_('s2', default=None):
        Or(
            None,
            dict,
            And(str, Use(lambda n: get_template(n, from_cat_data=False))),
            error='optional.qd.dissociate.s2 expects a string or a dictionary'
        ),

    Optional_("xyn_pre_opt", default=True): bool,
})

#: Schema for validating the ``['optional']['qd']['optimize']`` block.
qd_opt_schema: Schema = Schema({
    Optional_('use_ff', default=False):
        bool,

    Optional_('keep_files', default=True):
        And(bool, error='optional.qd.opt.keep_files expects a boolean'),

    # The job type for the first half of the optimization
    Optional_('job1', default=_get_amsjob):
        Or(
            And(
                And(type, lambda n: issubclass(n, Job), Use(val_job_type)),
                error=('optional.qd.opt.job1 expects a type object '
                       'that is a subclass of plams.Job')
            ),
            And(
                str, Use(str_to_job_type),
                error='optional.qd.opt.job1 expects a string that is a valid plams.Job alias'
            ),
            error='optional.qd.opt.job1 expects a string or a type object'
        ),

    # The job settings for the first half of the optimization
    Optional_('s1', default=_qd_opt_s1_default.copy()):
        Or(
            dict,
            And(str, Use(lambda n: get_template(n, from_cat_data=False))),
            error='optional.qd.opt.s1 expects a string or a dictionary'
        ),

    # The job type for the second half of the optimization
    Optional_('job2', default=_get_amsjob):
        Or(
            None,
            And(
                And(type, lambda n: issubclass(n, Job), Use(val_job_type)),
                error=('optional.qd.opt.job2 expects a type object '
                       'that is a subclass of plams.Job')
            ),
            And(
                str, Use(str_to_job_type),
                error='optional.qd.opt.job2 expects a string that is a valid plams.Job alias'
            ),
            error='optional.qd.opt.job2 expects a string or a type object'
        ),

    # The job settings for the second half of the optimization
    Optional_('s2', default=_qd_opt_s2_default.copy()):
        Or(
            None,
            dict,
            And(str, Use(lambda n: get_template(n, from_cat_data=False))),
            error='optional.qd.opt.s2 expects a string or a dictionary'
        )
})

#: Schema for validating the ``['optional']['ligand']['cosmo-rs']`` block.
crs_schema: Schema = Schema({
    # Delete files after the calculations are finished
    Optional_('keep_files', default=True):
        And(bool, error='optional.ligand.cosmo-rs.keep_files expects a boolean'),

    # The job type for constructing the COSMO surface
    Optional_('job1', default=_get_amsjob):
        Or(
            And(
                And(type, lambda n: issubclass(n, Job), Use(val_job_type)),
                error=('optional.ligand.cosmo-rs.job1 expects a type object '
                       'that is a subclass of plams.Job')
            ),
            And(
                str, Use(str_to_job_type),
                error=('optional.ligand.cosmo-rs.job1 expects a string '
                       'that is a valid plams.Job alias')
            ),
            error='optional.ligand.cosmo-rs.job1 expects a string or a type object'
        ),

    # The settings for constructing the COSMO surface
    Optional_('s1', default=_crs_s1_default.copy()):
        Or(
            dict,
            And(str, Use(lambda n: get_template(n, from_cat_data=False))),
            error='optional.ligand.cosmo-rs.s1 expects a string or a dictionary'
        ),

    Optional_('job2', default=_get_crsjob):  # The job type for the actual COSMO-RS calculation
        Or(
            And(
                And(type, lambda n: issubclass(n, Job), Use(val_job_type)),
                error=('optional.ligand.cosmo-rs.job2 expects a type object '
                       'that is a subclass of plams.Job')
            ),
            And(
                str, Use(str_to_job_type),
                error=('optional.ligand.cosmo-rs.job2 expects a string '
                       'that is a valid plams.Job alias')
            ),
            error='optional.ligand.cosmo-rs.job2 expects a string or a type object'
        ),

    # The settings for the actual COSMO-RS calculation
    Optional_('s2', default=_crs_s2_default.copy()):
        Or(
            dict,
            And(str, Use(lambda n: get_template(n, from_cat_data=False))),
            error='optional.ligand.cosmo-rs.s2 expects a string or a dictionary'
        )
})

#: Schema for validating the ``['optional']['qd']['activation_strain']`` block.
asa_schema: Schema = Schema({
    Optional_('use_ff', default=False):
        bool,

    Optional_('md', default=False):
        bool,

    Optional_('dump_csv', default=False):
        bool,

    Optional_('iter_start', default=500):
        And(val_index, lambda n: n.__index__() >= 0, Use(__index__)),

    Optional_('el_scale14', default=1.0):
        And(val_float, Use(float)),

    Optional_('lj_scale14', default=1.0):
        And(val_float, Use(float)),

    # Delete files after the calculations are finished
    Optional_('keep_files', default=True):
        And(bool, error='optional.qd.activation_strain.keep_files expects a boolean'),

    Optional_('distance_upper_bound', default=np.inf):
        Or(
            And(str, lambda n: 'inf' in n.lower(), Use(lambda n: np.inf)),
            And(val_float, lambda n: float(n) > 0, Use(float))
        ),

    Optional_('shift_cutoff', default=True):
        bool,

    Optional_('k', default=20):
        And(val_int, lambda n: int(n) > 0, Use(int)),

    # The job type for constructing the COSMO surface
    Optional_('job1', default=None):
        Or(
            None,
            And(
                And(type, lambda n: issubclass(n, Job), Use(val_job_type)),
                error=('optional.qd.activation_strain.job1 expects a type object '
                       'that is a subclass of plams.Job')
            ),
            And(
                str, Use(str_to_job_type),
                error=('optional.qd.activation_strain.job1 expects a string '
                       'that is a valid plams.Job alias')
            ),
            error='optional.qd.activation_strain.job1 expects a string or a type object'
        ),

    # The settings for constructing the COSMO surface
    Optional_('s1', default=None):
        Or(
            None,
            dict,
            And(str, Use(lambda n: get_template(n, from_cat_data=False))),
            error='optional.qd.activation_strain.s1 expects a string or a dictionary'
        ),
})

#: Schema for validating the ``['optional']['qd']['multi_ligand']`` block.
multi_ligand_schema: Schema = Schema({
    'ligands':
        And(
            abc.Collection,
            lambda n: not isinstance(n, str),
            lambda n: all(isinstance(i, str) for i in n),
            Use(tuple)
        ),

    Optional_('anchor', default=None):
        Or(
            None,
            And(abc.Collection,
                lambda n: not isinstance(n, str),
                lambda n: len(set(n)) == len(n),
                Use(lambda n: to_tuple(n, func=to_atnum)))
        ),

    # Alias for `optional.qd.multi_ligand.anchor`
    Optional_('dummy', default=None):
        Or(
            None,
            And(abc.Collection,
                lambda n: not isinstance(n, str),
                lambda n: len(set(n)) == len(n),
                Use(lambda n: to_tuple(n, func=to_atnum)))
        ),

    Optional_('f', default=None):
        Or(
            None,
            And(
                abc.Collection,
                lambda n: all(val_float(i) for i in n),
                lambda n: all(float(i) > 0 for i in n),
                lambda n: 0 < sum(float(i) for i in n) <= 1,
                Use(lambda n: to_tuple(n, func=float))
            )
        ),

    Optional_('mode', default='uniform'):
        And(str, lambda n: n.lower() in {'uniform', 'random', 'cluster'}, Use(str.lower)),

    Optional_('start', default=None):
        Or(
            None,
            And(val_index, Use(__index__))
        ),

    Optional_('follow_edge', default=False):
        bool,

    Optional_('cluster_size', default=1):
        Or(
            And(val_int, lambda n: int(n) > 0, Use(int)),
            And(abc.Collection,
                lambda n: all(val_int(i) for i in n),
                lambda n: all(int(i) > 0 for i in n),
                Use(lambda n: to_tuple(n, func=int)))
        ),

    Optional_('weight', default=lambda: str_to_func('np.exp(-x)')):
        Or(
            And(Callable, Use(str_to_func)),
            And(str, lambda n: 'x' in n, Use(str_to_func))
        ),

    Optional_('randomness', default=None):
        Or(
            None,
            And(val_float, lambda n: 0 <= float(n) <= 1, Use(float))
        )
})


#: Schema for validating the ``['optional']['ligand']['cdft']`` block.
cdft_schema: Schema = Schema({
    # Delete files after the calculations are finished
    Optional_('keep_files', default=True):
        And(bool, error='optional.ligand.cdft.keep_files expects a boolean'),

    # The Job type for the final geometry optimization
    Optional_('job1', default=lambda n: ADFJob):
        Or(
            And(
                And(type, lambda n: issubclass(n, Job), Use(val_job_type)),
                error=('optional.ligand.cdft.job1 expects a type object '
                       'that is a subclass of plams.Job')
            ),
            And(
                str, Use(str_to_job_type),
                error=('optional.ligand.cdft.job1 expects a string '
                       'that is a valid plams.Job alias')
            ),
        ),

    # The Job Settings for the final geometry optimization
    Optional_('s1', default=Settings):
        Or(
            None,
            dict,
            And(str, Use(lambda n: get_template(n, from_cat_data=False))),
            error='optional.ligand.cdft.s1 expects a string or a dictionary'
        ),
})



"""A module for multiple compound attachment and export of the .xyz files.

Index
-----
.. currentmodule:: CAT.dye.addlig
.. autosummary::
    add_ligands
    export_dyes
    sa_scores

API
---
.. autofunction:: add_ligands
.. autofunction:: export_dyes
.. autofunction:: sa_scores

"""

import os
import gzip
import math
import pickle
from typing import Iterator, Iterable, Collection, Optional, Dict
from os.path import join, exists
from itertools import chain

import numpy as np
from scm.plams import read_molecules, Molecule
from scm.plams.interfaces.molecule.rdkit import to_rdmol
from rdkit.Chem import rdMolDescriptors, FindMolChiralCenters, Mol

from nanoutils import PathType
from CAT.logger import logger
from CAT.attachment.dye import label_lig, label_core, substitution
from CAT.attachment.substitution_symmetry import del_equiv_structures

__all__ = ['add_ligands', 'export_dyes', 'sa_scores']

# flake8: noqa: N806


def add_ligands(core_dir: str,
                ligand_dir: str,
                min_dist: float = 1.2,
                n: int = 1,
                symmetry: Collection[str] = ()) -> Iterator[Molecule]:
    """Add ligand(s) to one core.

    Parameters
    ----------
    core_dir: str
        Name of directory where core coordinates are located
    ligand_dir: str
        Name of directory where ligands coordinates are located
    min_dist: float
        Criterion for the minimal interatomic distances
    n: int
        Number of substitutions
    symmetry: tuple[str]
        Keywords for substitution symmetry for deleting equivalent molecules

    Returns
    -------
    Iterator[Molecule]
        New structures that are containg core and lingad fragments

    """
    # Validate the symmetry
    valid_sym = {'linear', 'triangle', 'D2h'}
    if not valid_sym.issuperset(symmetry):
        raise ValueError(f"Invalid 'symmetry' value: {symmetry!r}")

    input_ligands = list(read_molecules(ligand_dir).values())
    input_cores = list(read_molecules(core_dir).values())

    # Reading the comment section and labeling substitution atom and numbering the ligands
    label_lig(input_ligands)

    # Reading the comment section and labeling substitution atoms
    for core in input_cores:
        core.guess_bonds()
        label_core(core)

    # Generate structures by combining ligands and cores
    mono = substitution(input_ligands, input_cores, min_dist)
    new_mols = []
    new_mols.append(mono)
    if n > 1:
        for i in range(1, n):
            poli = substitution(input_ligands, mono, min_dist)
            new_mols.append(poli)
            mono = poli

    if not symmetry:
        pass
    elif 'linear' in symmetry:
        new_mols[1] = del_equiv_structures(new_mols[1], 'linear')
    elif 'triangle' in symmetry:
        new_mols[2] = del_equiv_structures(new_mols[2], 'triangle')
    elif 'D2h' in symmetry:
        new_mols[3] = del_equiv_structures(new_mols[3], 'D2h')

    # Combine and flatten all new molecules into a generator
    return chain.from_iterable(new_mols)


def export_dyes(mol_list: Iterable[Molecule],
                new_dir: str = 'new_molecules',
                err_dir: str = 'err_molecules',
                min_dist: float = 1.2) -> None:
    """Exports molecular coordinates to .xyz files.

    Parameters
    ----------
    mol_list: Iterable[Molecule]
        List of molecules
    new_dir: str
        Name of the durectory to place new structures
    err_dir: str
        Name of the directory for molecules that do not fullfill min_dist criterion
    min_dist: float
        Criterion for the minimal interatomic distances

    """
    if not exists(new_dir):
        os.makedirs(new_dir)
    if not exists(err_dir):
        os.makedirs(err_dir)

    # Export molecules to folders depending on the minimum core/ligand distance
    for mol in mol_list:
        name = mol.properties.name
        mol_distance = mol.properties.min_distance

        if mol_distance > min_dist:
            mol.write(join(new_dir, f'{name}.xyz'))
        else:
            mol.write(join(err_dir, f'err_{name}.xyz'))
            logger.warning(
                f"{name}: \n Minimal distance {mol_distance} A is smaller than {min_dist} A"
            )

# The functions '_compute_SAS' and 'SA_scores' as well as data set 'SA_score.pkl.gz'
# are copied and adapted from:
###########################################################################
#    Title: MolGAN: An implicit generative model for small molecular graphs
#    Authors: De Cao, Nicola and Kipf, Thomas
#    Date: 2018
#    Availability: https://github.com/nicola-decao/MolGAN
###########################################################################


def _compute_sas(mol: Mol, sa_model: Dict[int, float]) -> float:
    fp = rdMolDescriptors.GetMorganFingerprint(mol, 2)
    fps = fp.GetNonzeroElements()
    score1 = 0.
    nf = 0
    # for bitId, v in fps.items():
    for bitId, v in fps.items():
        nf += v
        sfp = bitId
        score1 += sa_model.get(sfp, -4) * v
    score1 /= nf

    # features score
    nAtoms = mol.GetNumAtoms()
    nChiralCenters = len(FindMolChiralCenters(
        mol, includeUnassigned=True))
    ri = mol.GetRingInfo()
    nSpiro = rdMolDescriptors.CalcNumSpiroAtoms(mol)
    nBridgeheads = rdMolDescriptors.CalcNumBridgeheadAtoms(mol)
    nMacrocycles = 0
    for x in ri.AtomRings():
        if len(x) > 8:
            nMacrocycles += 1

    sizePenalty = nAtoms ** 1.005 - nAtoms
    stereoPenalty = math.log10(nChiralCenters + 1)
    spiroPenalty = math.log10(nSpiro + 1)
    bridgePenalty = math.log10(nBridgeheads + 1)
    macrocyclePenalty = 0.

    # ---------------------------------------
    # This differs from the paper, which defines:
    # macrocyclePenalty = math.log10(nMacrocycles+1)
    # This form generates better results when 2 or more macrocycles are present
    if nMacrocycles > 0:
        macrocyclePenalty = math.log10(2)

    score2 = 0. - sizePenalty - stereoPenalty - spiroPenalty - bridgePenalty - macrocyclePenalty

    # correction for the fingerprint density
    # not in the original publication, added in version 1.1
    # to make highly symmetrical molecules easier to synthetise
    score3 = 0.
    if nAtoms > len(fps):
        score3 = math.log(float(nAtoms) / len(fps)) * .5

    sascore = score1 + score2 + score3

    # need to transform "raw" value into scale between 1 and 10
    min = -4.0
    max = 2.5
    sascore = 11. - (sascore - min + 1) / (max - min) * 9.
    # smooth the 10-end
    if sascore > 8.:
        sascore = 8. + math.log(sascore + 1. - 9.)
    if sascore > 10.:
        sascore = 10.0
    elif sascore < 1.:
        sascore = 1.0

    return sascore


def _load_sa_model(filename: PathType) -> Dict[int, float]:
    sa_score = pickle.load(gzip.open(filename))
    return {j: i0 for i0, *i in sa_score for j in i}


def sa_scores(mols: Iterable[Molecule], filename: Optional[PathType] = None) -> np.ndarray:
    """Calculate the synthetic accessibility score for all molecules in **mols**."""
    sa_model = _load_sa_model(filename) if filename is not None else {}
    rdmols = (to_rdmol(mol) for mol in mols)

    try:
        count = len(mols)  # type: ignore
    except TypeError:
        count = -1
    iterator = (_compute_sas(mol, sa_model) for mol in rdmols)
    return np.fromiter(iterator, dtype=float, count=count)



"""A module for holding the :class:`WorkFlow` class.

Index
-----
.. currentmodule:: CAT.workflows.workflow
.. autosummary::
    WorkFlow

API
---
.. autoclass:: WorkFlow
    :members:

"""

import os
import operator
import threading
from shutil import rmtree
from pathlib import Path
from collections import abc
from typing import (
    Optional, Union, Dict, Hashable, MutableMapping, TypeVar, Iterable, Container, Tuple, Callable,
    Any, List, Type, Mapping, TYPE_CHECKING, cast, ClassVar, ContextManager
)

import numpy as np
import pandas as pd

import rdkit
import qmflows
from rdkit.Chem.AllChem import UFFGetMoleculeForceField as UFF  # noqa: N814
from scm.plams import finish, Settings, Molecule
from scm.plams.core.basejob import Job
from assertionlib import AbstractDataClass, NDRepr

from .key_map import MOL, OPT
from .workflow_dicts import WORKFLOW_TEMPLATE, _TemplateMapping
from ..utils import restart_init, parallel_init, JOB_MAP
from ..logger import logger
from ..settings_dataframe import SettingsDataFrame

if TYPE_CHECKING:
    from dataCAT import Database
    from dataCAT.database import JobRecipe
else:
    Database = 'dataCAT.Database'
    JobRecipe = 'dataCAT.database.JobRecipe'

__all__ = ['WorkFlow']

NDRepr.repr_SettingsDataFrame = NDRepr.repr_DataFrame  # type: ignore
aNDRepr = NDRepr()

T = TypeVar('T')


def _return_true(value: object) -> bool:
    """Return :data:`True`."""
    return True


def _lt_0(value) -> bool:
    """Return if **value** is smaller than ``0``."""
    return value < 0


def pop_and_concatenate(mapping: MutableMapping[str, T], base_key: object,
                        filter_func: Callable[[Any], bool] = _return_true) -> Tuple[T, ...]:
    """Take a key and :meth:`pop<dict.pop>` all values from **mapping**.

    The popping will continue as long as :code:`base_key + str(i)` is available in the mapping,
    where ``i`` is an :class:`int` larger than 1.
    The value if ``i`` will start from 1 and increase by `+ 1` every iteration.

    Examples
    --------
    .. code:: python

        >>> from CAT.workflows.workflow import pop_and_concatenate

        >>> mapping: dict = {
        ...     'job1': 1,
        ...     'job2': 2,
        ...     'job3': 3,
        ...     'final_key': True
        ... }

        >>> base_key = 'job'
        >>> value_tuple: tuple = pop_and_concatenate(mapping, base_key)
        >>> print(value_tuple)
        (1, 2, 3)

        >>> print(mapping)
        {'final_key': True}

    Parameters
    ----------
    mapping : :data:`MutableMapping`
        A dictionary or other mutable mapping.

    base_key : :data:`Hashable<typing.Hashable>`
        The base key which will be appended with successively increasing integers.

    filter_func : :data:`Callable<typing.Callable>`
        A callable for truth-testing each extracted **mapping** value.
        Values returning ``False`` will not be added to the to-be returned :class:`tuple`.

    Returns
    -------
    :class:`tuple`
        A tuple with values popped from **mapping**.

    """
    i = 1
    ret: List[T] = []
    while True:
        key = f'{base_key}{i}'
        try:
            value = mapping.pop(key)
        except KeyError:
            return tuple(ret)
        else:
            if filter_func(value):
                ret.append(value)
            i += 1


OptionalJobType = Union[None, Type[Job], Iterable[Optional[Type[Job]]]]
OptionalSettings = Union[None, Settings, Iterable[Optional[Settings]]]


class WorkFlow(AbstractDataClass):
    """A workflow manager.

    Examples
    --------
    Typical usage example:

    .. code:: python

        >>> import pandas as pd

        # Prepare workflow parameters
        >>> df = pd.DataFrame(...)  # doctest: +SKIP
        >>> settings = Settings()

        >>> def fancy_df_func(df, **kwargs):
        ...     pass

        # Create the workflow
        >>> workflow = WorkFlow.from_template(settings, name='asa')  # doctest: +SKIP
        >>> print(workflow)  # doctest: +SKIP
        WorkFlow(
            name       = 'asa',
            db         = None,
            read       = False,
            write      = False,
            overwrite  = False,
            path       = '.',
            keep_files = True,
            jobs       = None,
            settings   = None
        )

        # Run the workflow
        >>> idx = workflow.from_db(df)  # doctest: +SKIP
        >>> workflow(fancy_df_func, df, index=idx)  # doctest: +SKIP

        # Export all workflow results
        >>> job_recipe = workflow.get_recipe()  # doctest: +SKIP
        >>> workflow.to_db(df, job_recip=job_recipe)  # doctest: +SKIP

    """

    #: Map a name to a workflow template.
    _WORKFLOW_TEMPLATES: ClassVar[Mapping[str, _TemplateMapping]] = WORKFLOW_TEMPLATE

    #: A context manager for supressing Pandas :exc:`SettingwithCopyWarning`.
    _SUPRESS_SETTINGWITHCOPYWARNING: ClassVar[ContextManager[None]] = pd.option_context(
        'mode.chained_assignment', None
    )

    # Get-only properties

    @property
    def template(self) -> Mapping[str, Tuple[str, ...]]:
        """Get :attr:`WorkFlow._WORKFLOW_TEMPLATES` [:attr:`WorkFlow.name`] [``"template"``]."""
        return self._WORKFLOW_TEMPLATES[self.name]['template']

    @property
    def mol_type(self) -> str:
        """Get :attr:`WorkFlow._WORKFLOW_TEMPLATES` [:attr:`WorkFlow.name`] [``"mol_type"``]."""
        return self._WORKFLOW_TEMPLATES[self.name]['mol_type']

    @property
    def description(self) -> str:
        """Get :attr:`WorkFlow._WORKFLOW_TEMPLATES` [:attr:`WorkFlow.name`] [``"description"``]."""
        return self._WORKFLOW_TEMPLATES[self.name]['description']

    @property
    def import_columns(self) -> Mapping[Tuple[str, str], Any]:
        """Get :attr:`WorkFlow._WORKFLOW_TEMPLATES` [:attr:`WorkFlow.name`] [``"import_columns"``]."""  # noqa: E501
        return self._WORKFLOW_TEMPLATES[self.name]['import_columns']

    @property
    def export_columns(self) -> Tuple[Tuple[str, str], ...]:
        """Get :attr:`WorkFlow._WORKFLOW_TEMPLATES` [:attr:`WorkFlow.name`] [``"export_columns"``]."""  # noqa: E501
        return self._WORKFLOW_TEMPLATES[self.name]['export_columns']

    @property
    def dtype(self) -> np.dtype:
        """Get :attr:`WorkFlow._WORKFLOW_TEMPLATES` [:attr:`WorkFlow.name`] [``"dtype"``]."""
        return self._WORKFLOW_TEMPLATES[self.name]['dtype']

    # Getter and setter properties

    @property
    def read(self) -> bool:
        """Get or set :attr:`WorkFlow.read`.

        Setting accepts either a boolean or a container that may
        or may not contain :attr:`WorkFlow.mol_type` as value.

        """
        return self._read

    @read.setter
    def read(self, value: Union[bool, Container[str]]) -> None:
        try:
            self._read: bool = bool(self.db) and self.mol_type in value  # type: ignore
        except TypeError:  # value is not a container
            self._read = bool(value)

    @property
    def write(self) -> bool:
        """Get or set :attr:`WorkFlow.write`.

        Setting accepts either a boolean or a container that may
        or may not contain :attr:`WorkFlow.mol_type` as value.

        """
        return self._write

    @write.setter
    def write(self, value: Union[bool, Container[str]]) -> None:
        try:
            self._write: bool = bool(self.db) and self.mol_type in value
        except TypeError:  # value is not a container
            self._write = bool(value)

    @property
    def overwrite(self) -> bool:
        """Get or set :attr:`WorkFlow.overwrite`.

        Setting accepts either a boolean or a container that may
        or may not contain :attr:`WorkFlow.mol_type` as value.

        """
        return self._overwrite

    @overwrite.setter
    def overwrite(self, value: Union[bool, Container[str]]) -> None:
        try:
            self._overwrite: bool = bool(self.db) and self.mol_type in value
        except TypeError:  # value is not a container
            self._overwrite = bool(value)

    @property
    def jobs(self) -> Tuple[Optional[Type[Job]], ...]:
        """Get or set :attr:`WorkFlow.jobs`.

        Setting accepts either a |plams.Job| type, ``None`` or
        an iterable containing one (or both) of the aforementioned objects.

        """
        return self._jobs

    @jobs.setter
    def jobs(self, value: Union[None, Type[Job], Iterable[Optional[Type[Job]]]]) -> None:
        if isinstance(value, type):
            self._jobs: Tuple[Optional[Type[Job]], ...] = (value,)
        else:
            self._jobs = (None,) if value is None else tuple(value)

    @property
    def settings(self) -> Tuple[Optional[Settings], ...]:
        """Get or set :attr:`WorkFlow.settings`.

        Setting accepts either a |plams.Settings| instance, ``None`` or
        an iterable containing one (or both) of the aforementioned objects.

        """
        return self._settings

    @settings.setter
    def settings(self, value: Union[None, Settings, Iterable[Optional[Settings]]]) -> None:
        if isinstance(value, Settings):
            self._settings: Tuple[Optional[Settings], ...] = (value,)
        else:
            self._settings = (None,) if value is None else tuple(value)

    # Methods and magic methods

    def __init__(self, name: str,
                 db: Optional[Database] = None,
                 read: Union[bool, Container[str]] = False,
                 write: Union[bool, Container[str]] = False,
                 overwrite: Union[bool, Container[str]] = False,
                 path: Union[None, str, 'os.PathLike[str]'] = None,
                 keep_files: bool = True,
                 read_template: bool = True,
                 jobs: OptionalJobType = None,
                 settings: OptionalSettings = None,
                 thread_safe: bool = False,
                 **kwargs: Any) -> None:
        """Initialize a :class:`WorkFlow` instance; see also :meth:`Workflow.from_template`."""
        super().__init__()

        if name not in self._WORKFLOW_TEMPLATES:
            raise ValueError(f"Invalid value for the 'name' parameter: {name!r}\n"
                             f"Allowed values: {list(self._WORKFLOW_TEMPLATES.keys())!r}")

        self.name = name
        self.db = db

        self.read = cast(bool, read)
        self.write = cast(bool, write)
        self.overwrite = cast(bool, overwrite)

        self.path: Union[str, 'os.PathLike[str]'] = path if path is not None else os.getcwd()
        self.keep_files = keep_files
        self.thread_safe = thread_safe
        self.read_template = read_template
        self.jobs = cast(Tuple[Optional[Type[Job]], ...], jobs)
        self.settings = cast(Tuple[Optional[Settings], ...], settings)

        for k, v in kwargs.items():
            if hasattr(self, k):
                raise AttributeError(f"An attribute by the name of {k!r} already exists")
            setattr(self, k, v)

    @AbstractDataClass.inherit_annotations()
    def _str_iterator(self):
        iterator = super()._str_iterator()
        return ((k.strip('_'), v) for k, v in iterator)

    # TODO: Ensure that func allways returns an array-like object (no iterators)
    def __call__(self, func: Callable, df: pd.DataFrame,
                 index: Union[slice, pd.Series] = slice(None),
                 columns: Optional[List[Hashable]] = None,
                 no_loc: bool = False, **kwargs) -> None:
        r"""Initialize the workflow.

        Parameters
        ----------
        func : :data:`Callable<typing.Callable>`
            A callable object which will recieve **df**, all :class:`WorkFlow` instance
            attributes and ***kwargs** as arguments.
            The callable is expected to conduct some operation and export the results **dfs'**
            :attr:`WorkFlow.import_columns` columns.

        df : :class:`pandas.DataFrame`
            A DataFrame with molecules and results.

        index : :class:`slice` or :class:`pandas.Series` [:class:`bool`]
            An object for slicing the rows of **df** (*i.e.* a :attr:`pandas.DataFrame.index`).

        columns : :class:`list` [:data:`Hashable<typing.Hashable>`], optional
            An object for slicing the columns of **df** (*i.e.* :attr:`pandas.DataFrame.columns`).
            The output of **func** will be fed into :code:`df[columns]`.
            If ``None``, use :attr:`WorkFlow.import_columns` instead.

        no_loc : :class:`bool`
            If ``True``, substitute :meth:`pandas.DataFrame.loc` for
            :meth:`pandas.DataFrame.__setitem__`.
            Usefull to prevent Panda's rather aggressive typecasting when passing object arrays.
            Note that **no_loc** and **index** cannot be simultaneously specified.

        \**kwargs : :data:`Any<typing.Any>`
            Optional keyword arguments for **func**.

        See Also
        --------
        :meth:`Workflow.from_db`:
            Returns a value for the **idx_slice** parameter.

        """
        # Prepare slices
        if no_loc and index != slice(None):
            raise ValueError
        slice1 = index, MOL
        slice2 = index, list(self.import_columns.keys()) if columns is None else columns

        # Run the workflow
        logger.info(f"Starting {self.description}")
        with PlamsInit(path=self.path, folder=self.name,
                       thread_safe=self.thread_safe), self._SUPRESS_SETTINGWITHCOPYWARNING:
            self_vars = {k.strip('_'): v for k, v in vars(self).items()}
            value = func(df.loc[slice1], columns=columns, **self_vars, **kwargs)

            if not isinstance(value, abc.Iterator):
                try:  # This can fail for `np.flexible`-based dtypes
                    assert np.any(value)
                except (ValueError, TypeError):
                    pass
                except AssertionError:
                    return

            if no_loc:
                for k, v in zip(slice2[1], value):
                    df[k] = v
            else:
                try:
                    df.loc[slice2] = value
                except ValueError as ex:
                    logger.debug(f"df = {aNDRepr.repr(df)}")
                    logger.debug(f"index = {aNDRepr.repr(slice2[0])}")
                    logger.debug(f"columns = {aNDRepr.repr(slice2[1])}")
                    logger.debug(f"value = {aNDRepr.repr(value)}")
                    raise ex
        logger.info(f"Finishing {self.description}\n")

    def from_db(self, df: pd.DataFrame, inplace: bool = True, get_mol: bool = False,
                columns: Optional[Mapping] = None) -> Union[slice, pd.Series]:
        """Ensure that all required keys are present in **df** and import from the database.

        Returns a :class:`pandas.index` with all to-be updated rows, as based on how many
        previous results were imported from :attr:`WorkFlow.db`.
        If no results were pulled from :attr:`WorkFlow.db` (or :attr:`WorkFlow.overwrite` is
        ``True``), then return :code:`slice(None)`.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A DataFrame with molecules and results.

        inplace : :class:`bool`
            If ``True``, perform an inplace update of the Cartesian coordinates of all molecules
            rather than importing new molecules.

        get_mol : :class:`bool`
            If ``False`` do *not* try to import molecules from the database.

        columns : :class:`dict` [:data:`Hashable<typing.Hashable>`, :class:`object`], optional
            An dictionary whose keys will be used for slicing the columns of **df**
            (*i.e.* :attr:`pandas.DataFrame.columns`).
            The dictionary valeus are used as fill values if a key belongs
            to a to-be created column.
            Recommended values are :data:`numpy.nan`, ``None``, ``-1`` and/or ``False``.
            If ``None``, use :attr:`WorkFlow.import_columns` instead.

        Returns
        -------
        :class:`pandas.Series` [:class:`bool`] or :class:`slice`
            A Series for slicing a part of **df** or  a :class:`slice` object for
            slicing the entirety of **df** (*i.e.* :code:`slice(0, None`).

        """
        # Add all necasary keys to **df**
        import_columns: Mapping = self.import_columns if columns is None else columns
        for key, value in import_columns.items():
            if key not in df:
                df[key] = value

        if not self.read:  # Nothing to see here, move along
            return slice(None)

        # Import from the database
        with self._SUPRESS_SETTINGWITHCOPYWARNING:
            mol_list = self.db.from_csv(df, database=self.mol_type,
                                        get_mol=get_mol, inplace=inplace)
            if not inplace:  # mol_list is an actual sequence instead of None
                df[MOL] = mol_list

        # Return a new DataFrame slice based on previously calculated results
        if self.overwrite:
            return slice(None)
        else:
            keys = list(import_columns.keys())
            return self._isnull(df, keys).any(axis=1)

    def to_db(self, df: pd.DataFrame, status: Optional[str] = None,
              job_recipe: Optional[dict] = None,
              index: Union[slice, pd.Series] = slice(None),
              columns: Optional[Dict[Hashable, Any]] = None) -> None:
        """Export results to the database.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A DataFrame with molecules and results.

        status : :class:`str`, optional
            Whether or not **df** contains structures resulting from a geometry optimization.

        job_recipe : :class:`dict`
            A (nested) dictionary with the used job settings.

        index : :class:`slice` or :class:`pandas.Series` [:class:`bool`]
            An object for slicing the rows of **df** (*i.e.* :attr:`pandas.DataFrame.index`).

        columns : :class:`list` [:data:`Hashable<typing.Hashable>`], optional
            An object for slicing the columns of **df** (*i.e.* :attr:`pandas.DataFrame.columns`).
            If ``None``, use :attr:`WorkFlow.export_columns` instead.

        """
        # Dont export any settings columns if job_recipe is None
        # No job recipe == no settings to export anyway
        _export_columns = self.export_columns if columns is None else columns
        if job_recipe is None:
            export_columns = [i for i in _export_columns if i[0] != 'settings']
        else:
            export_columns = list(_export_columns)

        # Set the optimization status of the molecules to True
        if status == 'optimized':
            df.loc[index, OPT] = False

        # Write results to the database
        if self.write:
            with self._SUPRESS_SETTINGWITHCOPYWARNING:
                self.db.update_csv(
                    df, index,
                    database=self.mol_type,
                    columns=export_columns,
                    overwrite=self.overwrite,
                    job_recipe=job_recipe,
                    status=status,
                )

        # Remove the PLAMS results directories
        if not self.keep_files:
            name = self.name if not self.thread_safe else f'{self.name}.{threading.get_ident()}'
            rmtree(Path(self.path) / name)

    @classmethod
    def from_template(cls, settings: Union[Settings, SettingsDataFrame], name: str) -> 'WorkFlow':
        """Construct a :class:`WorkFlow` instance from a |plams.Settings| object.

        Parameters
        ----------
        settings : |plams.Settings|
            A Settings instance with all CAT settings.
            Certain values are extracted from **settings** based on the supplied template
            (see **name**).

        name : :class:`str`
            The name of the settings template.

        See Also
        --------
        :attr:`WorkFlow._WORKFLOW_TEMPLATES`
            A dictionary with all available template names (*i.e.* its keys).

        """
        # Extract the settings object from the SettingsDataFrame
        if isinstance(settings, SettingsDataFrame):
            settings = settings.settings

        kwargs: Dict[str, Any] = {'name': name}

        # Raise a KeyError if a key cannot be found
        with Settings.suppress_missing():
            try:  # Extract the correct template
                template = cls._WORKFLOW_TEMPLATES[name]['template']
            except KeyError as ex:
                err = (f"Invalid value for the 'name' parameter: {name!r}\n"
                       f"Allowed values: {', '.join(repr(k) for k in cls._WORKFLOW_TEMPLATES)}")
                raise ValueError(err) from ex

            # Create a dictionary with keyword arguments
            for k, v in template.items():
                kwargs[k] = settings.get_nested(v)

        # Post process all jobs and job settings
        kwargs['jobs'] = pop_and_concatenate(kwargs, 'job')
        kwargs['settings'] = pop_and_concatenate(kwargs, 's')
        return cls.from_dict(kwargs)

    def get_recipe(self) -> Dict[Tuple[str], JobRecipe]:
        """Create a recipe for :meth:`WorkFlow.to_db`."""
        settings_names = [i[1:] for i in self.export_columns if i[0] == 'settings']
        uff_fallback = {
            'key': f'RDKit_{rdkit.__version__}', 'value': f'{UFF.__module__}.{UFF.__name__}'
        }

        ret: Dict[Tuple[str], JobRecipe] = Settings()
        for name, job, settings in zip(settings_names, self.jobs, self.settings):
            # job is None, *i.e.* it's an RDKit UFF optimziation
            if job is None:
                ret[name].update(uff_fallback)  # type: ignore
                continue

            settings = Settings(settings)
            if self.read_template:  # Update the settings using a QMFlows template
                template = qmflows.geometry['specific'][self.type_to_string(job)].copy()
                settings.soft_update(template)
            ret[name]['key'] = job
            ret[name]['value'] = settings
        return ret

    @staticmethod
    def _isnull(df: pd.DataFrame, columns: list) -> pd.DataFrame:
        """A more expansive version of the :func:`pandas.isnull` function.

        :class:`int` series now also return ``True`` if smaller than ``0`` and :class:`bool`
        series are simply inverted.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            A DataFrame.

        columns : :class:`list`
            A list of column keys from **df**.

        """
        dtype_dict: Dict[np.dtype, Callable] = {
            np.dtype(bool): operator.invert,
            np.dtype(int): _lt_0,
            np.dtype(float): pd.isnull,
            np.dtype(object): pd.isnull
        }

        ret = pd.DataFrame(index=df.index)
        for key, series in df[columns].items():
            func = dtype_dict.get(series.dtype, pd.isnull)
            ret[key] = func(series)
        return ret

    @staticmethod
    def type_to_string(job: Union[Job, Type[Job]]) -> Optional[str]:
        """Turn a :class:`type` instance into a :class:`str`.

        Accepts one of the following |plams.Job| subclasses:
            * ``ADFJob``
            * ``AMSJob``
            * ``DiracJob``
            * ``Cp2kJob``
            * ``GamessJob``
            * ``ORCAJob``
            * ``CRSJob``

        Parameters
        ----------
        job : :class:`type` [|plams.Job|] or |plams.Job|
            A PLAMS Job type or instance.

        Returns
        -------
        :class:`str`, optional
            Returns either ``None`` or an item pulled from :data:`._job_dict`.

        """
        if not isinstance(job, type):
            job = type(job)  # Convert a class instance into a class

        try:
            return JOB_MAP[job]
        except KeyError as ex:
            logger.error(f"No default settings available for type: '{job.__class__.__name__}'")
            logger.debug(f'{ex.__class__.__name__}: {ex}', exc_info=True)
            return None

    @staticmethod
    def pop_job_settings(mol_list: Iterable[Molecule], key: str = 'job_path') -> List[List[str]]:
        """Take a list of molecules and pop and return all references to **key**.

        Parameters
        ----------
        mol_list : :data:`Iterable<typing.Iterable>` [|plams.Molecule|]
            An iterable consisting of PLAMS molecules.
            For this method to be effective they should contain a property by the name of **key**:
            a list of strings represnting paths to .in files.

        key : :data:`Hashable<typing.Hashable>`
            The to-be popped key in each molecule in **mol_list**.

        Returns
        -------
        :class:`list` [:class:`list` [:class:`str`]]
            A nested list of strings popped from **mol_list**.

        """
        ret = []
        for mol in mol_list:
            try:
                ret.append(mol.properties.pop(key))
            except KeyError:
                ret.append([])
            mol.properties[key] = []
        return ret


class PlamsInit(ContextManager[None]):
    """A context manager for calling :func:`.restart_init` and |plams.finish|."""

    def __init__(self, path: Union[str, 'os.PathLike[str]'],
                 folder: Union[str, 'os.PathLike[str]'],
                 hashing: str = 'input',
                 *, thread_safe: bool = False):
        self.path = path
        self.folder = folder
        self.hashing = hashing
        self.thread_safe = thread_safe

    def __enter__(self) -> None:
        """Enter the context manager; call :func:`.restart_init`."""
        if not self.thread_safe:
            restart_init(self.path, self.folder, self.hashing)
        else:
            parallel_init(self.path, self.folder, self.hashing)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the context manager; call |plams.finish|."""
        finish()



"""A set of :class:`WorkFlow` related modules."""

from . import key_map
from .key_map import *
from .workflow_dicts import WORKFLOW_TEMPLATE
from .workflow import WorkFlow

__all__ = ['WorkFlow', 'WORKFLOW_TEMPLATE']
__all__ += key_map.__all__

# flake8: noqa: F403,F401



"""A module designed for running Jobs.

Index
-----
.. currentmodule:: CAT.jobs
.. autosummary::
    get_main_molecule
    _xyz_to_mol
    get_energy
    _get_name
    pre_process_settings
    retrieve_results
    job_single_point
    job_geometry_opt
    job_freq

API
---
.. automethod:: get_main_molecule
.. autofunction:: _xyz_to_mol
.. automethod:: get_energy
.. autofunction:: _get_name
.. autofunction:: pre_process_settings
.. autofunction:: retrieve_results
.. autofunction:: job_single_point
.. autofunction:: job_geometry_opt
.. autofunction:: job_freq

"""

from shutil import rmtree
from typing import (Optional, Type)
from os.path import join

import numpy as np

from scm.plams.core.basejob import Job
from scm.plams import (Molecule, Settings, Results, config, add_to_class, ResultsError,
                       ADFJob, AMSJob, Units, Cp2kResults)

import qmflows

from .logger import (logger, log_start, log_succes, log_fail, log_copy)
from .thermo_chem import get_thermo
from .utils import type_to_string

__all__ = ['job_single_point', 'job_geometry_opt', 'job_freq']


@add_to_class(Cp2kResults)
def get_main_molecule(self) -> Optional[Molecule]:
    try:
        filename = self.files['cp2k-pos-1.xyz']
        return _xyz_to_mol(filename)
    except TypeError:
        pass

    iterator = (join(self.job.path, f) for f in self.files if '.xyz' in f)
    for filename in iterator:
        return _xyz_to_mol(filename)
    raise ResultsError(f'Failed to retrieve main molecule from {self.job.name}')


def _xyz_to_mol(filename: str) -> Molecule:
    """Grab the last geometry from an .xyz file and return it as a :class:`Molecule` instance."""
    with open(filename, 'r') as f:
        atom_count = int(next(f))
        for line_count, _ in enumerate(f, 2):
            pass

    mol_count = line_count // (2 + atom_count)
    return Molecule(filename, geometry=mol_count)


@add_to_class(Cp2kResults)
def get_energy(self, index: int = -1, unit: str = 'Hartree') -> float:
    """Return the energy of the last occurence of ``'ENERGY| Total FORCE_EVAL'`` in the output."""
    energy_str = self.grep_output('ENERGY| Total FORCE_EVAL')[index]
    energy = float(energy_str.rsplit(maxsplit=1)[1])
    return Units.convert(energy, 'Hartree', unit)


def _get_name(name: str) -> str:
    manager = config.default_jobmanager
    if name in manager.names:
        number = str(1 + manager.names[name]).zfill(manager.settings.counter_len)
        return f'{name}.{number}'
    return name


def pre_process_settings(mol: Molecule, s: Settings,
                         job_type: Type[Job], template_name: str) -> Settings:
    """Update all :class:`Settings`, **s**, with those from a QMFlows template (see **job**)."""
    ret = Settings()
    type_key = type_to_string(job_type)
    ret.input = getattr(qmflows, template_name)['specific'][type_key].copy()
    ret.update(s)

    if job_type is AMSJob:
        mol.properties.pop('charge', None)
        # ret.input.ams.system.BondOrders._1 = adf_connectivity(mol)
        if 'uff' not in s.input:
            ret.input.ams.system.charge = int(sum(
                at.properties.get('charge', 0) for at in mol
            ))

    elif job_type is ADFJob:
        mol.properties.pop('charge', None)
        if not ret.input.charge:
            ret.input.charge = int(sum(at.properties.get('charge', 0) for at in mol))
    return ret


JOB_PRESETS = frozenset({
    'geometry optimization', 'frequency analysis', 'single point', 'MD calculation'
})


def retrieve_results(mol: Molecule, results: Results, job_preset: str) -> None:
    """Unpack the :class:`results` from a PLAMS-facilitated calculation.

    Performs an inplace update of **results**:

        * The Cartesian coordinates of :attr:`Results.job.molecule` are updated.
        * Energies and frequencies are added to :attr:`Results.job.molecule.properties`.
        *

    Paramaters
    ----------
    mol : |plams.Molecule|_
        A PLAMS molecule.

    results : |plams.Results|_
        A PLAMS :class:`Results` instance.

    job_preset : str
        The name of a job preset from :mod:`CAT.jobs`.
        Accepted values are ``"single point"``, ``"geometry optimization"`` or
        ``"frequency analysis"``.

    """
    if job_preset not in JOB_PRESETS:
        raise ValueError(f'Invalid value for job_preset: {job_preset!r}')

    # Unpack arguments
    job = results.job
    name = job.name

    # Define more arguments
    freq = np.empty(1)
    nan_dict = {
        'frequencies': np.nan,
        'energy': {'E': np.nan, 'H': np.nan, 'S': np.nan, 'G': np.nan}
    }

    try:
        if job.status in {'failed', 'crashed'}:
            raise _get_results_error(results)

        # Read all relevant results
        energy = mol.properties.energy.E = results.get_energy(unit='kcal/mol')
        if job_preset in ('geometry optimization', 'frequency analysis'):
            mol_new = results.get_main_molecule() or Molecule()
            mol.from_array(mol_new)

        if job_preset == 'frequency analysis':
            freq = mol.properties.frequencies = results.get_frequencies()
            energy = mol.properties.energy = get_thermo(mol, freq, energy)

        # Evaluate all results
        if not (energy and isinstance(freq, np.ndarray)):
            raise _get_results_error(results)
        log_succes(job, mol, job_preset, name)

    except Exception as ex:  # Failed to retrieve results
        if job_preset == 'geometry optimization':
            mol.properties.is_opt = False
        mol.properties.soft_update(nan_dict)
        log_fail(job, mol, job_preset, name)
        logger.debug(f'{ex.__class__.__name__}: {ex}', exc_info=True)

    else:
        if job_preset == 'geometry optimization':
            mol.properties.is_opt = True
        if job.status != 'copied':
            return None

        # results.job is a copy from a previously run job.
        # Exchange the attributes of results.job wth those of the revious job
        job_old = config.default_jobmanager.hashes[job._hash]()
        log_copy(job, mol, job_preset, name, job_old)
        rmtree(job.path)
        for key, value in vars(job_old).items():
            if key != 'molecule':
                setattr(job, key, value)
    return None


def _get_results_error(results: Results) -> ResultsError:
    """Raise a :exc:`ResultsError` with the content of ``results['$JN.err']`` as error mesage."""
    filename = results['$JN.err']
    with open(filename, 'r') as f:
        iterator = (i.rstrip('\n') for i in f)
        for item in iterator:
            if item:
                return ResultsError(item)
        else:
            return ResultsError()


@add_to_class(Job)
def _finalize(self):
    """Modified PLAMS :meth:`Job._finalize` method.

    1. All references to the :func:`scm.plams.log` function are removed.
    2. A $JN.hash file is created which contains the hash of this instance.

    Gather the results of the job execution and organize them.
    This method collects steps 9-12 from :ref:`job-life-cycle`.
    Should not be overridden.

    """
    if config.preview is False:
        self.results.collect()
        self.results.finished.set()
        if self.status != 'crashed':
            self.status = 'finished'
            self._log_status(3)
            if self.check():
                self.results._clean(self.settings.keep)
                self.postrun()
                self.status = 'successful'
                if self.settings.pickle:
                    self.pickle()
                    filename = join(self.path, f'{self.name}.hash')
                    with open(filename, 'w') as f:
                        try:  # Will raise a TypeError if rerun prevetion is disabled
                            f.write(self.hash())
                        except TypeError:
                            pass
            else:
                self.status = 'failed'
    else:
        self.status = 'preview'
        self.results.finished.set()
    self.results.done.set()

    if self.parent and self in self.parent:
        self.parent._notify()

    self._log_status(1)


@add_to_class(Molecule)
def job_single_point(self, job_type: Type[Job],
                     settings: Settings,
                     name: str = 'Single_point',
                     ret_results: bool = False,
                     read_template: bool = True) -> Optional[Results]:
    """Function for running an arbritrary jobs, extracting total energies.

    Paramaters
    ----------
    job_type : |Callable|_
        A type Callable of a class derived from :class:`Job`, e.g. :class:`AMSJob`
        or :class:`Cp2kJob`.

    settings : |plams.Settings|_
        The settings for **job**.

    name : str
        The name of **job**.

    ret_results : bool
        Whether or not the :class:`Results` instance should be returned or not.

    read_template : bool
        Whether or not to update **settings** using a QMFlows template or not.

    Returns
    -------
    |plams.Results|_
        Optional: If ``ret_results=True` return the :class:`Results` instance produced by this job.

    """
    # Grab the default settings for a specific job and update them with user provided settings
    _sp = 'singlepoint'
    s = settings if not read_template else pre_process_settings(self, settings, job_type, _sp)

    # Run the job; extract energies
    job = job_type(molecule=self, settings=s, name=name)
    _name = _get_name(name)
    log_start(job, self, 'single point', _name)

    results = job.run()
    retrieve_results(self, results, 'single point')

    inp_name = join(job.path, f'{job.name}.in')
    try:
        self.properties.job_path.append(inp_name)
    except TypeError:
        self.properties.job_path = [inp_name]

    # Return results
    if ret_results:
        return job.results
    return None


@add_to_class(Molecule)
def job_geometry_opt(self, job_type: Type[Job],
                     settings: Settings,
                     name: str = 'Geometry_optimization',
                     ret_results: bool = False,
                     read_template: bool = True) -> Optional[Results]:
    """Function for running an arbritrary jobs, extracting total energies and final geometries.

    Paramaters
    ----------
    job_type : |Callable|_
        A type Callable of a class derived from :class:`Job`, e.g. :class:`AMSJob`
        or :class:`Cp2kJob`.

    settings : |plams.Settings|_
        The settings for **job**.

    name : str
        The name of **job**.

    ret_results : bool
        Whether or not the :class:`Results` instance should be returned or not.

    read_template : bool
        Whether or not to update **settings** using a QMFlows template or not.

    Returns
    -------
    |plams.Results|_
        Optional: If ``ret_results=True` return the :class:`Results` instance produced by this job.

    """
    # Grab the default settings for a specific job and update them with user provided settings
    _geo = 'geometry'
    s = settings if not read_template else pre_process_settings(self, settings, job_type, _geo)

    # Run the job; extract geometries and energies
    job = job_type(molecule=self, settings=s, name=name)
    _name = _get_name(name)
    log_start(job, self, 'geometry optimization', _name)

    results = job.run()
    retrieve_results(self, results, 'geometry optimization')

    inp_name = join(job.path, f'{job.name}.in')
    try:
        self.properties.job_path.append(inp_name)
    except TypeError:
        self.properties.job_path = [inp_name]

    # Return results
    if ret_results:
        return job.results
    return None


@add_to_class(Molecule)
def job_md(self, job_type: Type[Job],
           settings: Settings,
           name: str = 'MD',
           ret_results: bool = False,
           read_template: bool = False) -> Optional[Results]:
    """Function for running an arbritrary jobs, extracting total energies and final geometries.

    Paramaters
    ----------
    job_type : |Callable|_
        A type Callable of a class derived from :class:`Job`, e.g. :class:`AMSJob`
        or :class:`Cp2kJob`.

    settings : |plams.Settings|_
        The settings for **job**.

    name : str
        The name of **job**.

    ret_results : bool
        Whether or not the :class:`Results` instance should be returned or not.

    read_template : bool
        Whether or not to update **settings** using a QMFlows template or not.

    Returns
    -------
    |plams.Results|_
        Optional: If ``ret_results=True` return the :class:`Results` instance produced by this job.

    """
    # Grab the default settings for a specific job and update them with user provided settings
    s = settings

    # Run the job; extract geometries and energies
    job = job_type(molecule=self, settings=s, name=name)
    _name = _get_name(name)
    log_start(job, self, 'MD calculation', _name)

    results = job.run()
    retrieve_results(self, results, 'MD calculation')

    inp_name = join(job.path, f'{job.name}.in')
    try:
        self.properties.job_path.append(inp_name)
    except TypeError:
        self.properties.job_path = [inp_name]

    # Return results
    if ret_results:
        return job.results
    return None


@add_to_class(Molecule)
def job_freq(self, job_type: Type[Job],
             settings: Settings,
             name: str = 'Frequency_analysis',
             opt: bool = True,
             ret_results: bool = False,
             read_template: bool = True) -> Optional[Results]:
    """Function for running an arbritrary Jobs.

    Extracts total energies, final geometries and
    thermochemical quantities derived from vibrational frequencies.

    Paramaters
    ----------
    job : |Callable|_
        A type Callable of a class derived from :class:`Job`, e.g. :class:`AMSJob`
        or :class:`Cp2kJob`.

    settings : |plams.Settings|_
        The settings for **job**.

    name : str
        The name of **job**.

    opt : bool
        Perform a geometry optimization (see :func:`.job_geometry_opt`) before calculating
        frequencies.

    ret_results : bool
        Whether or not the :class:`Results` instance should be returned or not.

    read_template : bool
        Whether or not to update **settings** using a QMFlows template or not.

    Returns
    -------
    |plams.Results|_
        Optional: If ``ret_results=True` return the :class:`Results` instance produced by this job.

    """
    # Preceed the frequency analysis with a geometry optimization
    if opt:
        self.job_geometry_opt(job_type, settings)

    # Grab the default settings for a specific job and update them with user provided settings
    s = settings if not read_template else pre_process_settings(self, settings, job_type, 'freq')

    # Run the job; extract geometries and (Gibbs free) energies
    job = job_type(molecule=self, settings=s, name=name)
    _name = _get_name(name)
    log_start(job, self, 'frequency analysis', _name)

    results = job.run()
    retrieve_results(self, results, 'frequency analysis')

    inp_name = join(job.path, f'{_name}.in')
    try:
        self.properties.job_path.append(inp_name)
    except TypeError:
        self.properties.job_path = [inp_name]

    # Return results
    if ret_results:
        return job.results
    return None



"""A module designed for attaching ligands to cores.

Index
-----
.. currentmodule:: CAT.attachment.ligand_attach
.. autosummary::
    init_qd_construction
    construct_mol_series
    _read_database
    _get_indices
    _get_df
    ligand_to_qd
    _get_rotmat1
    _get_rotmat2
    rot_mol
    rot_mol_angle
    array_to_qd
    sanitize_dim_2
    sanitize_dim_3

API
---
.. autofunction:: init_qd_construction
.. autofunction:: construct_mol_series
.. autofunction:: _read_database
.. autofunction:: _get_indices
.. autofunction:: _get_df
.. autofunction:: ligand_to_qd
.. autofunction:: _get_rotmat1
.. autofunction:: _get_rotmat2
.. autofunction:: rot_mol
.. autofunction:: rot_mol_angle
.. autofunction:: array_to_qd
.. autofunction:: sanitize_dim_2
.. autofunction:: sanitize_dim_3

"""

from typing import List, Tuple, Any, Optional, NoReturn, Union, Iterable
from collections import abc

import numpy as np
import pandas as pd
from scipy.spatial import cKDTree
from scipy.spatial.distance import cdist

from scm.plams import Molecule, Atom, Settings, MoleculeError
from assertionlib.ndrepr import aNDRepr

from .perp_surface import get_surface_vec
from ..mol_utils import get_index, round_coords  # noqa: F401
from ..utils import AnchorTup
from ..workflows import WorkFlow, HDF5_INDEX, MOL, OPT
from ..settings_dataframe import SettingsDataFrame
from ..data_handling import mol_to_file, WARN_MAP
from ..utils import AllignmentTup, AllignmentEnum

__all__ = ['init_qd_construction']


def init_qd_construction(ligand_df: SettingsDataFrame, core_df: SettingsDataFrame,
                         construct_qd: bool = True) -> SettingsDataFrame:
    """Initialize the quantum dot construction.

    Parameters
    ----------
    ligand_df : |CAT.SettingsDataFrame|_
        A dataframe of ligands.

    core_df : |CAT.SettingsDataFrame|_
        A dataframe of cores.

    construct_qd : :class:`bool`
        If ``False``, only construct and return the dataframe without filling it with molecules.

    Returns
    -------
    |CAT.SettingsDataFrame|_
        A dataframe of quantum dots.

    """
    # import pdb; pdb.set_trace()
    qd_df = _get_df(core_df.index, ligand_df.index, ligand_df.settings)
    qd_df[MOL] = None
    qd_df.sort_index(inplace=True)
    if not construct_qd:
        return qd_df

    workflow = WorkFlow.from_template(qd_df, name='qd_attach')
    workflow.keep_files = False

    # Pull from the database; push unoptimized structures
    idx = workflow.from_db(qd_df, inplace=False, get_mol=True)

    # Start the ligand optimization
    workflow(construct_mol_series, qd_df, columns=MOL, index=idx,
             core_df=core_df, ligand_df=ligand_df)
    workflow.to_db(qd_df, index=idx)

    # Export ligands to .xyz, .pdb, .mol and/or .mol format
    mol_format = qd_df.settings.optional.database.mol_format
    if mol_format and not qd_df.settings.optional.qd.optimize:
        path = workflow.path
        mol_to_file(qd_df.loc[idx, MOL], path, mol_format=mol_format)

    return qd_df


def construct_mol_series(qd_df: SettingsDataFrame, core_df: pd.DataFrame,
                         ligand_df: pd.DataFrame, path: str,
                         allignment: str = 'sphere', **kwargs: Any) -> pd.Series:
    """Construct a Series of new quantum dots."""
    def _get_mol(i, j, k, m):
        ij = i, j
        km = k, m
        return ligand_to_qd(core_df.at[ij, MOL], ligand_df.at[km, MOL], path, allignment=allignment)

    mol_list = [_get_mol(i, j, k, m) for i, j, k, m in qd_df.index]
    return pd.Series(mol_list, index=qd_df.index, name=MOL, dtype=object)


def _get_indices(mol: Molecule,
                 index: Tuple[str, str, str, str]) -> List[int]:
    """Return a list with the indices of all atoms in the core plus ligand anchor atoms.

    Ligand anchor atoms are furthermore marked with the properties.anchor attribute.

    Parameters
    ----------
    mol : |plams.Molecule|_
        A PLAMS molecule.

    index : |tuple|_ [|str|_]
        A tuple of 4 strings.

    Returns
    -------
    |list|_ [|int|_]
        A list of atomic indices.

    """
    # Collect the indices of the atoms in the core
    ret = []
    for i, at in enumerate(mol, 1):
        if at.properties.pdb_info.ResidueName == 'COR':
            ret.append(i)
        else:
            break

    # Extract the index (within the ligand) of the ligand anchor atom
    index = index[3]
    for j, _ in enumerate(index):
        try:
            k = index[j:] - 1
            break
        except ValueError:
            pass
    k += i - 1

    # Append and return
    ref_name = mol[k+1].properties.pdb_info.Name
    for i, at in enumerate(mol.atoms[k:], k+1):
        if at.properties.pdb_info.Name == ref_name:
            at.properties.anchor = True
            ret.append(i)
    return ret


def _get_df(core_index: pd.MultiIndex,
            ligand_index: pd.MultiIndex,
            settings: Settings) -> SettingsDataFrame:
    """Create and return a new quantum dot dataframe.

    Parameters
    ----------
    core_index : |pd.MultiIndex|_
        A multiindex of the cores.

    ligand_index : |pd.MultiIndex|_
        A multiindex of the ligands.

    settings : |plams.Settings|_
        A Settings intance extracted from the ligand or core dataframe.

    Returns
    -------
    |CAT.SettingsDataFrame|_
        An empty dataframe (*i.e.* filled with ``-1`` and ``False``) of quantum dots.

    """
    # Create the index
    idx_tups = [(i, j, k, l) for i, j in core_index for k, l in ligand_index]
    index = pd.MultiIndex.from_tuples(
        idx_tups, names=['core', 'core anchor', 'ligand smiles', 'ligand anchor']
    )

    # Create the collumns
    column_tups = [HDF5_INDEX, OPT]
    columns = pd.MultiIndex.from_tuples(column_tups, names=['index', 'sub index'])

    # Create and return the quantum dot dataframe
    data = {HDF5_INDEX: -1, OPT: False}
    return SettingsDataFrame(data, index=index, columns=columns, settings=settings)


def ligand_to_qd(
    core: Molecule,
    ligand: Molecule,
    path: str,
    allignment: AllignmentTup,
    idx_subset: "None | Iterable[int]" = None,
) -> Molecule:
    """Function that handles quantum dot (qd, *i.e.* core + all ligands) operations.

    Combine the core and ligands and assign properties to the quantom dot.

    Parameters
    ----------
    core : |plams.Molecule|_
        A core molecule.

    ligand : |plams.Molecule|_
        A ligand molecule.

    allignment : :class:`str`
        How the core vector(s) should be defined.
        Accepted values are ``"sphere"`` and ``"surface"``:

        * ``"sphere"``: Vectors from the core anchor atoms to the center of the core.
        * ``"surface"``: Vectors perpendicular to the surface of the core.

        Note that for a perfect sphere both approaches are equivalent.

    idx_subset : :class:`Iterable<collections.anc.Iterable>` [:class:`int`], optional
        An iterable with the (0-based) indices defining a subset of atoms in **core**.
        Only relevant in the construction of the convex hull when ``allignment=surface``.

    Returns
    -------
    |plams.Molecule|_
        A quantum dot consisting of a core molecule and *n* ligands

    """
    def get_name() -> str:
        core_name = core.properties.name
        anchor = str(qd[-1].properties.pdb_info.ResidueNumber - 1)
        lig_name = ligand.properties.name
        return f'{core_name}__{anchor}_{lig_name}'

    idx_subset_ = idx_subset if idx_subset is not None else ...

    # Define vectors and indices used for rotation and translation the ligands
    vec1 = np.array([-1, 0, 0], dtype=float)  # All ligands are already alligned along the X-axis
    idx = ligand.get_index(ligand.properties.dummies) - 1
    ligand.properties.dummies.properties.anchor = True

    # Attach the rotated ligands to the core, returning the resulting strucutre (PLAMS Molecule).
    anchor = sanitize_dim_2(core.properties.dummies)
    if allignment.kind == AllignmentEnum.SPHERE:
        vec2 = np.array(core.get_center_of_mass()) - anchor
        vec2 /= np.linalg.norm(vec2, axis=1)[..., None]
    elif allignment.kind == AllignmentEnum.SURFACE:
        vec2 = -get_surface_vec(np.array(core)[idx_subset_], anchor)
    else:
        raise ValueError(f"Unknown allignment kind: {allignment.kind}")
    if allignment.invert:
        vec2 *= -1

    # Rotate the ligands
    anchor_tup = ligand.properties.anchor_tup
    if anchor_tup.dihedral is None:
        lig_array = rot_mol(
            ligand, vec1, vec2, atoms_other=core.properties.dummies, core=core, idx=idx
        )
    else:
        lig_array = rot_mol(
            ligand, vec1, vec2, atoms_other=core.properties.dummies, core=core,
            idx=idx, anchor_tup=anchor_tup,
        )

    # Combine the ligands and core
    qd = core.copy()
    array_to_qd(ligand, lig_array, mol_out=qd)
    qd.round_coords()

    # Set properties
    qd.properties = Settings({
        'indices': [i for i, at in enumerate(qd, 1) if
                    at.properties.pdb_info.ResidueName == 'COR' or at.properties.anchor],
        'path': path,
        'name': get_name(),
        'job_path': [],
        'prm': ligand.properties.get('prm')
    })

    # Print and return
    _evaluate_distance(qd, qd.properties.name)
    return qd


def _get_dihedrals(mat1: np.ndarray, mat2: np.ndarray, vec3: np.ndarray) -> np.ndarray:
    """Get the dihedral angle between three vectors in radian."""
    v1v2 = np.cross(-mat1, mat2)
    v2v3 = np.cross(vec3, mat2)
    v1v2_v2v3 = np.cross(v1v2, v2v3)
    v2_norm_v2 = mat2 / np.linalg.norm(mat2, axis=1)[..., None]
    return np.arctan2(
        np.einsum("ij,ij->i", v1v2_v2v3, v2_norm_v2),
        np.einsum("ij,ij->i", v1v2, v2v3),
    )


def _get_rotmat1(vec1: np.ndarray, vec2: np.ndarray) -> np.ndarray:
    """Calculate the rotation matrix for rotating **vec1** to **vec2**.

    Returns a unit matrix if the length of one of the vectors is 0.

    Paramaters
    ----------
    vec1 : :math:`3` or :math:`n*3` |np.ndarray|_
        An array-like object representing a single or :math:`n` vectors of length 3.
        Represents the initial orientation.

    vec2 : :math:`3` or :math:`n*3` |np.ndarray|_
        An array-like object representing a single or :math:`n` vectors of length 3.
        Represents the final orientation.

    Returns
    -------
    :math:`n*3*3` |np.ndarray|_
        An array consisting of :math:`n` rotation matrices.

    """
    # Increase the vector dimensionality if required
    u_vec1 = np.array(vec1, dtype=float, ndmin=2, copy=False)
    u_vec2 = np.array(vec2, dtype=float, ndmin=2, copy=False)

    with np.errstate(divide='raise', invalid='raise'):
        # Return a unit matrix if either vec1 or vec2 is zero
        try:
            u_vec1 /= np.linalg.norm(u_vec1, axis=1)[:, None]
            u_vec2 /= np.linalg.norm(u_vec2, axis=1)[:, None]
        except FloatingPointError:  # i.e. division by zero
            shape = max(len(u_vec1), len(u_vec2)), 3, 3
            ret = np.zeros(shape, dtype=float)
            ret[:] = np.identity(3)
            return ret

    v1, v2, v3 = np.cross(u_vec1, u_vec2).T
    v0 = np.zeros(max(len(u_vec1), len(u_vec2)))
    m = np.array([[v0, -v3, v2],
                  [v3, v0, -v1],
                  [-v2, v1, v0]]).T

    with np.errstate(invalid='ignore'):
        if len(u_vec1) > 1 and len(u_vec2) > 1:
            ret = np.identity(3) + m + ((m@m).T / (1 + u_vec2.T@u_vec1).T[..., None]).T
        elif len(u_vec1) == 1:
            ret = np.identity(3) + m + ((m@m).T / (1 + u_vec2@u_vec1[0].T).T).T
        elif len(u_vec2) == 1:
            ret = np.identity(3) + m + ((m@m).T / (1 + u_vec2[0]@u_vec1.T).T).T
        else:
            raise ValueError('vec1 and vec2 expect 1- or 2-dimensional array-like objects; '
                             f'observed shapes: vec1: {np.asarray(vec1).shape} and vec2: '
                             f'{np.asarray(vec2).shape}')

    # Ensure that there are no nan values in the rotation matrix
    isnan = np.isnan(ret)
    if not isnan.any():
        return ret
    else:
        i = np.unique(np.nonzero(isnan)[0])
        ret[i] = -np.eye(3)
        return ret


def _get_rotmat2(
    vec: np.ndarray,
    step: float = 1/16,
    angle_vec: "None | np.ndarray" = None,
) -> np.ndarray:
    r"""Calculate the rotation matrix for rotating m vectors along their axes, each vector yielding :math:`k = (2 / step)` possible rotations.

    Paramaters
    ----------
    vec : :math:`3` or :math:`n*3` |np.ndarray|_
        An array-like object representing a single or :math:`n` vectors of length 3.

    step : float
        The rotation stepsize as fraction of :math:`2*/pi`.

    """  # noqa
    # Increase the vector dimensionality if required and create unit vectors
    v = np.array(vec, dtype=float, ndmin=2, copy=False)
    v /= np.linalg.norm(v, axis=1)[:, None]

    v1, v2, v3 = v.T
    zero = np.zeros(len(v))
    w = np.array([[zero, -v3, v2],
                  [v3, zero, -v1],
                  [-v2, v1, zero]]).T

    if angle_vec is None:
        step_range = np.pi * np.arange(0.0, 2.0, step)
        a1 = np.sin(step_range)[:, None, None, None]
        a2 = 1 - np.cos(step_range)[:, None, None, None]
    else:
        a1 = np.sin(angle_vec)[:, None, None]
        a2 = 1 - np.cos(angle_vec)[:, None, None]
    return np.identity(3) + a1 * w + a2 * w@w


def rot_mol(xyz_array: np.ndarray,
            vec1: np.ndarray,
            vec2: np.ndarray,
            idx: int = 0,
            atoms_other: Optional[np.ndarray] = None,
            core: Optional[np.ndarray] = None,
            bond_length: Optional[int] = None,
            step: float = 1/16,
            dist_to_self: bool = True,
            ret_min_dist: bool = False,
            anchor_tup: "None | AnchorTup" = None) -> np.ndarray:
    r"""Rotate **xyz_array**.

    Paramaters
    ----------
    xyz_array : :math:`m*n*3` |np.ndarray|_ or :math:`m` |list|_ [|plams.Molecule|_]:
        A 3D array-like object or list of :math:`m` PLAMS molecules consisting of :math:`n` atoms.
        This array represents the to-be rotated molecules.

    vec1 : :math:`n*3` |np.ndarray|_:
        One or multiple vectors of length 3.
        Represents the initial orientation of **xyz_array**.

    vec2 : :math:`n*3` |np.ndarray|_:
        One or multiple vectors of length 3.
        Represents the final orientation of **xyz_array**.
        Also used as a rotation axis.

    idx : int:
        An atomic index or sequence of atomic indices.
        Used for translational purposes (see **atoms_other** and **bond_length**).
        Expects 0-based indices.

    atoms_other : :math:`n*3` |np.ndarray|_:
        Optional: A 2D array of Cartesian coordinates.
        If not ``None`` then **xyz_array** will be translated to ensure
        :code:`xyz_array[:, idx] == atoms_other`.

    bond_length : float
        Optional: One or multiple floats representing bond lengths.
        Performs a translation of :code:`xyz_array[:, idx]` along **vec2** by **bond_length**.

    step : float
        The rotation stepsize as fraction of :math:`2*\pi`.

    dist_to_self : bool
        pass

    Returns
    -------
    :math:`m*n*3` |np.ndarray|_:
        The rotated **xyz_array**.

    """
    # Turn all arguments into numpy arrays with appropiate dimensions
    xyz = sanitize_dim_3(xyz_array)
    vec1 = sanitize_dim_2(vec1)
    vec2 = sanitize_dim_2(vec2)

    # Define slices
    idx1 = np.arange(len(xyz)), idx
    idx2 = np.arange(len(xyz)), slice(None), idx

    # Translate xyz[idx] to the origin and rotate
    xyz -= xyz[idx1][..., None, :]
    rotmat1 = _get_rotmat1(vec1, vec2)
    xyz = xyz@rotmat1

    # Code-path for fixed-angle dihedrals
    if anchor_tup is not None:
        i, j, *_ = anchor_tup.anchor_idx
        vec3 = xyz[:, i] - xyz[:, j]
        dihedral_vec = _get_dihedrals(vec3, vec2, vec1)
        dihedral_vec -= anchor_tup.dihedral
        rotmat2 = _get_rotmat2(vec2, angle_vec=dihedral_vec)
        xyz = np.matmul(xyz, rotmat2, out=xyz)

        at_other = sanitize_dim_2(atoms_other)
        xyz += (at_other - xyz[:, idx])[:, None]
        return xyz

    # Create all k possible rotations of all m ligands
    rotmat2 = _get_rotmat2(vec2, step=step)
    xyz = np.swapaxes(xyz@rotmat2, 0, 1)
    if atoms_other is None:
        return xyz

    # Translate the the molecules in xyz_array
    at_other = sanitize_dim_2(atoms_other)
    xyz += (at_other[..., None, :] - xyz[idx2])[..., None, :]
    if bond_length is not None:
        bond_length = np.asarray(bond_length)
        mult = (bond_length / np.linalg.norm(vec2, axis=1))[:, None]
        xyz -= (vec2 * mult)[:, None, None, :]

    # Returns the conformation of each molecule that maximizes the inter-moleculair distance
    # Or return all conformations if dist_to_self = False and atoms_other = None
    return rotation_check_kdtree(xyz, at_other, core, ret_min_dist=ret_min_dist)


def rotation_check_kdtree(
    xyz: np.ndarray,
    core_anchor: np.ndarray,
    core: np.ndarray,
    k: int = 10,
    ret_min_dist: bool = False,
):
    """Perform the rotation check using SciPy's :class:`cKDTree<scipy.spatial.cKDTree.

    Parameters
    ----------
    xyz : :math:`(m, n, l, 3)` :class:`numpy.ndarray`
        A 4D array of Cartesian coordinates representing the :math:`n` rotameters of :math:m`
        ligands with :math:`l` atoms each.

    at_other : :math:`(m, 3)` :class:`numpy.ndarray`
        A 2D array with the Cartesian of neighbouring atoms.

    k : :class:`int`
        The number of to-be considered neighbours when performing the rotation check.

    See Also
    --------
    :meth:`cKDTree.query<scipy.spatial.cKDTree.query>`
        Query the kd-tree for nearest neighbors.

    """
    a, b, c, d = xyz.shape
    ret = np.empty((a, c, d), order='F')
    min_dist = np.empty(len(ret))
    distance_upper_bound = _get_distance_upper_bound(core_anchor)

    # Shrink down the core, keep the 6 atoms closest to the core-anchor
    core = np.asarray(core)
    tree = cKDTree(core)
    _, idx = tree.query(core_anchor, k=range(2, 7), distance_upper_bound=10)
    idx_ravel = np.unique(idx.ravel())
    idx_ravel = idx_ravel[idx_ravel != len(core)]
    core_subset = core[idx_ravel]

    at_other = core_subset
    for i, ar in enumerate(xyz):
        tree = cKDTree(at_other)
        dist, _ = tree.query(ar.reshape(b*c, d), k=k, distance_upper_bound=distance_upper_bound)
        dist.shape = b, c, k

        weighted_dist = np.exp(-dist).sum(axis=(1, 2))
        idx_min = weighted_dist.argmin()
        at_other = np.concatenate((at_other, ar[idx_min]))

        ret[i] = ar[idx_min]
        min_dist[i] = weighted_dist[idx_min]

    if ret_min_dist:
        return ret, min_dist
    return ret


def _evaluate_distance(mol: np.ndarray, name: str,
                       threshold: float = 1.0,
                       action: str = 'warn') -> Union[None, NoReturn]:
    """Eavluate all the distance matrix of **xyz3D** and perform **action** when distances are below **threshold**."""  # noqa
    try:
        action_func = WARN_MAP[action]
    except KeyError as ex:
        raise ValueError("'action' expected either 'warn', 'raise' or 'ignore'; "
                         f"observed value: {action!r}") from ex
    else:
        if action == 'ignore':
            return None

    xyz = np.asarray(mol)

    tree = cKDTree(xyz)
    dist, idx = tree.query(xyz, k=[2])
    dist = dist.T[0]
    idx = idx.T[0]

    bool_ar = dist < threshold
    if bool_ar.any():
        _idx2 = np.stack([np.arange(len(idx)), idx]).T
        _idx2 += 1
        _idx2.sort(axis=1)

        idx2 = np.unique(_idx2[bool_ar], axis=0)
        n = len(idx2)

        exc = MoleculeError(
            f"\nEncountered >= {n} unique atom-pairs at a distance shorter than "
            f"{threshold} Angstrom in {name!r}:\n{aNDRepr.repr(idx2.T)}"
        )
        action_func(exc)


def _get_distance_upper_bound(at_other: np.ndarray, r_min: float = 5.0,
                              r_max: float = 10.0) -> float:
    """Construct an estimate for **distance_upper_bound** based on the mean nearest-neighbour distance in **at_other**.

    The to-be returned value is clipped, if necessary, by **r_min** and **r_max**.

    """  # noqa
    tree = cKDTree(at_other)
    dist, _ = tree.query(at_other, k=2, distance_upper_bound=10.0)
    dist = dist.T[1]
    dist[dist == np.inf] = np.nan

    # Check if distance_upper_bound=10.0 is too small; try again with the full dist matrix if so
    if np.isnan(dist).all():
        dist = cdist(at_other, at_other)
        np.fill_diagonal(dist, np.inf)
        r = dist.min(axis=0).mean()
    else:
        r = np.nanmean(dist)

    # Clip and return
    return np.clip(r, r_min, r_max)


def rot_mol_angle(xyz_array: np.ndarray,
                  vec1: np.ndarray,
                  vec2: np.ndarray,
                  idx: int = 0,
                  atoms_other: Optional[np.ndarray] = None,
                  bond_length: Optional[float] = None) -> np.ndarray:
    """Rotate one or more molecules (**xyz_array**) from **vec1** to **vec2**.

    Paramaters
    ----------
    xyz_array : :math:`m*n*3` |np.ndarray|_ or :math:`m` |list|_ [|plams.Molecule|_]:
        A 3D array-like object or list of :math:`m` PLAMS molecules consisting of :math:`n` atoms.
        This array represents the to-be rotated molecules.

    vec1 : :math:`n*3` |np.ndarray|_:
        One or multiple vectors of length 3.
        Represents the initial orientation of **xyz_array**.

    vec2 : :math:`n*3` |np.ndarray|_:
        One or multiple vectors of length 3.
        Represents the final orientation of **xyz_array**.

    idx : int:
        An atomic index or sequence of atomic indices.
        Used for translational purposes (see **atoms_other** and **bond_length**).
        Expects 0-based indices.

    atoms_other : :math:`n*3` |np.ndarray|_:
        Optional: A 2D array of Cartesian coordinates.
        If not ``None`` then **xyz_array** will be translated to ensure
        :code:`xyz_array[:, idx] == atoms_other`.

    bond_length : float
        Optional: One or multiple floats representing bond lengths.
        Performs a translation of :code:`xyz_array[:, idx]` along **vec2** by **bond_length**.

    Returns
    -------
    :math:`m*n*3` |np.ndarray|_:
        The rotated **xyz_array**.

    """
    # Turn all arguments into numpy arrays with appropiate dimensions
    _xyz_array = sanitize_dim_3(xyz_array)
    _vec1 = sanitize_dim_2(vec1)
    _vec2 = sanitize_dim_2(vec2)

    # Rotate and translate all n ligands; readjust bond lengths if bond_length is set
    rotmat = _get_rotmat1(_vec1, _vec2)
    ret = _xyz_array@rotmat
    _idx = slice(0, None), idx

    # Translate the the molecules in ret
    if atoms_other is not None:
        atoms_other = sanitize_dim_2(atoms_other)
        ret += (atoms_other - ret[_idx])[:, None, :]
        if bond_length is not None:
            mult = (np.asarray(bond_length) / np.linalg.norm(vec2, axis=1))[:, None]
            ret -= (vec2 * mult)[:, None, :]

    # Return a m*n*3 array
    return ret


def array_to_qd(mol: Molecule, xyz_array: np.ndarray,
                mol_out: Optional[Molecule] = None) -> Optional[List[Molecule]]:
    """Create :math:`n` copies of **mol** and update their Cartesian coordinates with **xyz_array**.

    Parameters
    ----------
    mol : |plams.Molecule|_
        A template PLAMS molecule consisting of :math:`n` atoms.

    xyz_array : :math:`m*n*3` |np.ndarray|_:
        A 3D array-like object representing the cartesian coordinates of
        :math:`m` molecules each with :math:`n` atoms.

    mol_out : |plams.Molecule|_
        Optional: If not ``None`` merge the to-be returned molecules with **mol_out**.

    Returns
    -------
    |list|_ [|Molecule|_]
        Optional: if **mol_out** = ``None`` return a list **mol** copies, each with its
        Cartesian coordinates replaced with a set of coordinates from **xyz_array**.

    """
    mol_list = []
    xyz_array = sanitize_dim_3(xyz_array)
    if mol_out is None:
        start = 2
    else:
        start = 1 + mol_out[-1].properties.get('pdb_info', {}).get('ResidueNumber', 1)

    for i, xyz in enumerate(xyz_array, start=start):
        mol_cp = mol.copy()
        mol_cp.from_array(xyz)
        for at in mol_cp:
            at.properties.pdb_info.ResidueNumber = i
        mol_list.append(mol_cp)

    # return a list of molecules or merge the list with an existing molecule
    if mol_out is None:
        return mol_list
    else:
        for m in mol_list:
            mol_out.add_molecule(m)
        return None


def _is_sequence(item) -> bool:
    return isinstance(item, abc.Sequence)


def _is_atom(item) -> bool:
    return isinstance(item, Atom)


def _is_mol(item) -> bool:
    return isinstance(item, Molecule)


def _is_atom_sequence(item) -> bool:
    return _is_mol(item) or (_is_sequence(item) and _is_atom(item[-1]))


def _is_mol_sequence(item) -> bool:
    return _is_sequence(item) and _is_atom_sequence(item[-1])


def sanitize_dim_2(value: Any) -> np.ndarray:
    """Convert a PLAMS atom or sequence of :math:`n` PLAMS atoms into a :math:`n*3` array.

    Parameters
    ----------
    value : object
        The to be parsed object.
        Acceptable types are:

        * A PLAMS :class:`Atom`.
        * A PLAMS :class:`Molecule`.
        * A Sequence consisting of PLAMS :class:`Atom`.
        * An array-like object with a dimensionality smaller than 2 and float-compatible elements.

    Returns
    -------
    :math:`n*3` |np.ndarray|_ [|np.float64|_]
        A 2D array consisting of floats.

    Raises
    ------
    ValueError
        Raised if dimensionality of the to-be returned array is higher than 2
        or the content of **arg** cannot be converted into an array of floats.

    """
    value = value.coords if isinstance(value, Atom) else value
    try:
        ret = np.array(value, dtype=float, copy=False)
    except TypeError:
        ret = Molecule.as_array(None, atom_subset=value)

    if ret.ndim < 2:
        ret.shape = (2 - ret.ndim) * (1,) + ret.shape
    elif ret.ndim > 2:
        raise ValueError(f"Failed to create a 2D array; observed dimensionality: {ret.ndim}")
    return ret


def sanitize_dim_3(value: Any, padding: float = np.nan) -> np.ndarray:
    """Convert a Molecule or sequence of :math:`m` molecules into an :math:`m*n*3` array.

    If necessary, the to-be returned array is padded with **padding** .

    Parameters
    ----------
    arg : object
        The to be parsed object.
        Acceptable types are:

        * A PLAMS :class:`Atom`
        * A PLAMS :class:`Molecule`
        * A (nested) Sequences consisting of PLAMS :class:`Atom`
        * A (nested) Sequences consisting of PLAMS :class:`Molecule`
        * An array-like object with a dimensionality smaller than 3 and float-compatible elements.

    padding : float
        A value used for padding the to-be returned array.
        Only relevant if **arg** consists of multiple molecule with different numbers of atoms.

    Returns
    -------
    :math:`m*n*3` |np.ndarray|_ [|np.float64|_]
        A 3D array consisting of floats.

    Raises
    ------
    ValueError
        Raised if dimensionality of the to-be returned array is higher than 3
        or the content of **arg** cannot be converted into an array of floats.

    """
    if _is_atom(value):
        return np.array(value.coords)[None, None, :]

    elif _is_atom_sequence(value):
        return Molecule.as_array(None, atom_subset=value)[None, :]

    elif _is_mol_sequence(value):
        max_at = max(len(mol) for mol in value)
        ret = np.full((len(value), max_at, 3), padding, order='F')
        for i, mol in enumerate(value):
            j = len(mol)
            ret[i, :j] = Molecule.as_array(None, atom_subset=mol)
        return ret

    else:
        ret = np.array(value, ndmin=3, dtype=float, copy=False)
        if ret.ndim > 3:
            raise ValueError(f"Failed to create a 3D array; observed dimensionality: {ret.ndim}")
        return ret



"""Placeholder."""

from .attachment import distribution
from .attachment.distribution import (
    distribute_idx as distribute_idx,
    uniform_idx as uniform_idx,
)

# flake8: noqa: E402

__all__ = distribution.__all__
__doc__ = distribution.__doc__
del distribution



"""Modules related to the importing, exporting and general handling of data."""

from .mol_import import read_mol, set_mol_prop
from .warn_map import WARN_MAP
from .mol_to_file import mol_to_file


__all__ = [
    'read_mol', 'set_mol_prop',
    'WARN_MAP',
    'mol_to_file'
]



#!/usr/bin/env python
"""Entry points for CAT.

Index
-----
.. currentmodule:: CAT.data_handling.input_parser
.. autosummary::
    extract_args
    main

API
---
.. autofunction:: extract_args
.. autofunction:: main

"""

import argparse
from os import getcwd
from os.path import join, exists
from typing import Optional, List

import yaml

from scm.plams import Settings

from CAT import base
from CAT.logger import logger
from CAT.utils import log_traceback_locals


def extract_args(args: Optional[List[str]] = None) -> Settings:
    """Extract and return all arguments."""
    input_file = args.YAML[0]
    if exists(input_file):
        pass
    elif exists(join(getcwd(), input_file)):
        input_file = join(getcwd(), input_file)
    else:
        input_file2 = join(getcwd(), input_file)
        raise FileNotFoundError(f'No file found at {input_file} or {input_file2}')

    with open(input_file, 'r') as file:
        return Settings(yaml.load(file, Loader=yaml.FullLoader))


def main(args: Optional[List[str]] = None) -> None:
    """Launch CAT via the command line."""
    parser = argparse.ArgumentParser(
        prog='CAT',
        usage='init_cat my_settings_file.yaml',
        description=('Description: This script initalizes the Compound Attachment Tool (CAT).')
    )

    parser.add_argument(
        'YAML', nargs=1, type=str, metavar='input.yaml',
        help='Required: A .yaml file with the settings for CAT'
    )

    try:
        args = parser.parse_args(args)
        base.prep(extract_args(args), return_mol=False)
    except Exception:
        log_traceback_locals(logger)
        raise
